nace_kode_map = {
    11100: {
        "branche": "Dyrkning af korn (undtagen ris), bælgfrugter og olieholdige frø",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    11200: {
        "branche": "Dyrkning af ris",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    11300: {
        "branche": "Dyrkning af grøntsager og meloner, rødder og rodknolde",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    11400: {
        "branche": "Dyrkning af sukkerrør",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    11500: {
        "branche": "Dyrkning af tobak",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    11600: {
        "branche": "Dyrkning af tekstilplanter",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    11900: {
        "branche": "Dyrkning af andre etårige afgrøder",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    12100: {
        "branche": "Dyrkning af druer",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    12200: {
        "branche": "Dyrkning af tropiske og subtropiske frugter",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    12300: {
        "branche": "Dyrkning af citrusfrugter",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    12400: {
        "branche": "Dyrkning af kernefrugter og stenfrugter",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    12500: {
        "branche": "Dyrkning af andre træfrugter, bær og nødder",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    12600: {
        "branche": "Dyrkning af olieholdige frugter",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    12700: {
        "branche": "Dyrkning af planter til fremstilling af drikkevarer",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    12800: {
        "branche": "Dyrkning af krydderiplanter, aromaplanter og lægeplanter",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    12900: {
        "branche": "Dyrkning af andre flerårige afgrøder",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    13000: {
        "branche": "Planteformering",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    14100: {
        "branche": "Avl af malkekvæg",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    14200: {
        "branche": "Avl af andet kvæg og bøfler",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    14300: {
        "branche": "Avl af heste og dyr af hestefamilien",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    14400: {
        "branche": "Avl af kameler og dyr af kamelfamilien",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    14500: {
        "branche": "Avl af får og geder",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    14610: {
        "branche": "Avl af smågrise",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    14620: {
        "branche": "Produktion af slagtesvin",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    14700: {
        "branche": "Fjerkræavl",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    14910: {
        "branche": "Kenneler",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    14920: {
        "branche": "Avl af pelsdyr mv.",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    15000: {
        "branche": "Blandet drift",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    16100: {
        "branche": "Serviceydelser i forbindelse med planteavl",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    16200: {
        "branche": "Serviceydelser i forbindelse med husdyravl",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    16300: {
        "branche": "Forarbejdning af afgrøder efter høst",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    16400: {
        "branche": "Forarbejdning af frø/sædekorn til udsæd",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    17000: {
        "branche": "Jagt, fældefangst og serviceydelser i forbindelse hermed",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 1000,
        "grp127hovedafdeling": "Landbrug og gartneri",
    },
    21000: {
        "branche": "Dyrkning af træer og andre skovbrugsaktiviteter",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 2000,
        "grp127hovedafdeling": "Skovbrug",
    },
    22000: {
        "branche": "Skovning",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 2000,
        "grp127hovedafdeling": "Skovbrug",
    },
    23000: {
        "branche": "Indsamling af vildtvoksende forstmateriale undtagen træer",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 2000,
        "grp127hovedafdeling": "Skovbrug",
    },
    24000: {
        "branche": "Serviceydelser til skovbrug",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 2000,
        "grp127hovedafdeling": "Skovbrug",
    },
    31100: {
        "branche": "Havfiskeri",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 3000,
        "grp127hovedafdeling": "Fiskeri",
    },
    31200: {
        "branche": "Ferskvandsfiskeri",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 3000,
        "grp127hovedafdeling": "Fiskeri",
    },
    32100: {
        "branche": "Havbrug",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 3000,
        "grp127hovedafdeling": "Fiskeri",
    },
    32200: {
        "branche": "Ferskvandsbrug",
        "grp10kode": 1,
        "grp10hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp19kode": "A",
        "grp19hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp36kode": "A",
        "grp36hovedafdeling": "Landbrug, skovbrug og fiskeri",
        "grp127kode": 3000,
        "grp127hovedafdeling": "Fiskeri",
    },
    51000: {
        "branche": "Indvinding af stenkul",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 8009,
        "grp127hovedafdeling": "Indvinding af grus og sten",
    },
    52000: {
        "branche": "Indvinding af brunkul",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 8009,
        "grp127hovedafdeling": "Indvinding af grus og sten",
    },
    61000: {
        "branche": "Indvinding af råolie",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 6000,
        "grp127hovedafdeling": "Indvinding af olie og gas",
    },
    62000: {
        "branche": "Indvinding af naturgas",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 6000,
        "grp127hovedafdeling": "Indvinding af olie og gas",
    },
    71000: {
        "branche": "Brydning af jernmalm",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 8009,
        "grp127hovedafdeling": "Indvinding af grus og sten",
    },
    72100: {
        "branche": "Brydning af uran- og thoriummalme",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 8009,
        "grp127hovedafdeling": "Indvinding af grus og sten",
    },
    72900: {
        "branche": "Brydning af andre ikke-jernholdige metalmalme",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 8009,
        "grp127hovedafdeling": "Indvinding af grus og sten",
    },
    81100: {
        "branche": "Brydning af pynte- og bygningssten, kalksten, gips, kridt og skifer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 8009,
        "grp127hovedafdeling": "Indvinding af grus og sten",
    },
    81200: {
        "branche": "Grus- og sandgravning; indvinding af ler og kaolin",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 8009,
        "grp127hovedafdeling": "Indvinding af grus og sten",
    },
    89100: {
        "branche": "Indvinding af mineraler til fremstilling af kemiske produkter og gødningsstoffer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 8009,
        "grp127hovedafdeling": "Indvinding af grus og sten",
    },
    89200: {
        "branche": "Indvinding og agglomerering af tørv",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 8009,
        "grp127hovedafdeling": "Indvinding af grus og sten",
    },
    89300: {
        "branche": "Saltindvinding",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 8009,
        "grp127hovedafdeling": "Indvinding af grus og sten",
    },
    89900: {
        "branche": "Anden råstofindvinding i.a.n.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 8009,
        "grp127hovedafdeling": "Indvinding af grus og sten",
    },
    91000: {
        "branche": "Serviceydelser i forbindelse med indvinding af råolie og naturgas",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 9000,
        "grp127hovedafdeling": "Service til råstofindvinding",
    },
    99000: {
        "branche": "Serviceydelser i forbindelse med anden råstofindvinding",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "B",
        "grp19hovedafdeling": "Råstofindvinding",
        "grp36kode": "B",
        "grp36hovedafdeling": "Råstofindvinding",
        "grp127kode": 9000,
        "grp127hovedafdeling": "Service til råstofindvinding",
    },
    101110: {
        "branche": "Forarbejdning af svinekød",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10001,
        "grp127hovedafdeling": "Slagterier",
    },
    101190: {
        "branche": "Forarbejdning af andet kød",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10001,
        "grp127hovedafdeling": "Slagterier",
    },
    101200: {
        "branche": "Forarbejdning og konservering af fjerkrækød",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10001,
        "grp127hovedafdeling": "Slagterier",
    },
    101300: {
        "branche": "Produktion af kød- og fjerkrækødprodukter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10001,
        "grp127hovedafdeling": "Slagterier",
    },
    102010: {
        "branche": "Fremstilling af fiskemel",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10002,
        "grp127hovedafdeling": "Fiskeindustri",
    },
    102020: {
        "branche": "Forarbejdning og konservering af fisk, krebsdyr og bløddyr, undtagen fiskemel",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10002,
        "grp127hovedafdeling": "Fiskeindustri",
    },
    103100: {
        "branche": "Forarbejdning og konservering af kartofler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10005,
        "grp127hovedafdeling": "Anden fødevareindustri",
    },
    103200: {
        "branche": "Fremstilling af frugt- og grøntsagssaft",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10005,
        "grp127hovedafdeling": "Anden fødevareindustri",
    },
    103900: {
        "branche": "Anden forarbejdning og konservering af frugt og grøntsager",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10005,
        "grp127hovedafdeling": "Anden fødevareindustri",
    },
    104100: {
        "branche": "Fremstilling af olier og fedtstoffer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10005,
        "grp127hovedafdeling": "Anden fødevareindustri",
    },
    104200: {
        "branche": "Fremstilling af margarine o.l. spiselige fedtstoffer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10005,
        "grp127hovedafdeling": "Anden fødevareindustri",
    },
    105100: {
        "branche": "Mejerier samt ostefremstilling",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10003,
        "grp127hovedafdeling": "Mejerier",
    },
    105200: {
        "branche": "Fremstilling af konsumis",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10003,
        "grp127hovedafdeling": "Mejerier",
    },
    106100: {
        "branche": "Fremstilling af mølleriprodukter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10004,
        "grp127hovedafdeling": "Bagerier, brødfabrikker mv.",
    },
    106200: {
        "branche": "Fremstilling af stivelse og stivelsesprodukter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10004,
        "grp127hovedafdeling": "Bagerier, brødfabrikker mv.",
    },
    107110: {
        "branche": "Industriel fremstilling af brød; kager mv.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10004,
        "grp127hovedafdeling": "Bagerier, brødfabrikker mv.",
    },
    107120: {
        "branche": "Fremstilling af friske bageriprodukter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10004,
        "grp127hovedafdeling": "Bagerier, brødfabrikker mv.",
    },
    107200: {
        "branche": "Fremstilling af tvebakker og kiks; fremstilling af konserverede kager, tærter mv.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10004,
        "grp127hovedafdeling": "Bagerier, brødfabrikker mv.",
    },
    107300: {
        "branche": "Fremstilling af makaroni, nudler, couscous og lignende dejvarer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10004,
        "grp127hovedafdeling": "Bagerier, brødfabrikker mv.",
    },
    108100: {
        "branche": "Fremstilling af sukker",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10005,
        "grp127hovedafdeling": "Anden fødevareindustri",
    },
    108200: {
        "branche": "Fremstilling af kakao, chokolade og sukkervarer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10005,
        "grp127hovedafdeling": "Anden fødevareindustri",
    },
    108300: {
        "branche": "Forarbejdning af te og kaffe",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10005,
        "grp127hovedafdeling": "Anden fødevareindustri",
    },
    108400: {
        "branche": "Fremstilling af smagspræparater og krydderier",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10005,
        "grp127hovedafdeling": "Anden fødevareindustri",
    },
    108500: {
        "branche": "Fremstilling af færdigretter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10005,
        "grp127hovedafdeling": "Anden fødevareindustri",
    },
    108600: {
        "branche": "Fremstilling af homogeniserede produkter og diætmad",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10005,
        "grp127hovedafdeling": "Anden fødevareindustri",
    },
    108900: {
        "branche": "Fremstilling af andre fødevarer i.a.n.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10005,
        "grp127hovedafdeling": "Anden fødevareindustri",
    },
    109100: {
        "branche": "Fremstilling af færdige foderblandinger til landbrugsdyr",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10005,
        "grp127hovedafdeling": "Anden fødevareindustri",
    },
    109200: {
        "branche": "Fremstilling af færdige foderblandinger til kæledyr",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 10005,
        "grp127hovedafdeling": "Anden fødevareindustri",
    },
    110100: {
        "branche": "Destillation, rektifikation og blanding af alkohol",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 11000,
        "grp127hovedafdeling": "Drikkevareindustri",
    },
    110200: {
        "branche": "Fremstilling af vin af druer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 11000,
        "grp127hovedafdeling": "Drikkevareindustri",
    },
    110300: {
        "branche": "Fremstilling af cider og anden frugtvin",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 11000,
        "grp127hovedafdeling": "Drikkevareindustri",
    },
    110400: {
        "branche": "Fremstilling af andre ikke-destillerede gærede drikkevarer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 11000,
        "grp127hovedafdeling": "Drikkevareindustri",
    },
    110500: {
        "branche": "Fremstilling af øl",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 11000,
        "grp127hovedafdeling": "Drikkevareindustri",
    },
    110600: {
        "branche": "Fremstilling af malt",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 11000,
        "grp127hovedafdeling": "Drikkevareindustri",
    },
    110700: {
        "branche": "Fremstilling af læskedrikke; fremstilling af mineralvand og andet vand på flaske",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 11000,
        "grp127hovedafdeling": "Drikkevareindustri",
    },
    120000: {
        "branche": "Fremstilling af tobaksprodukter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CA",
        "grp36hovedafdeling": "Føde-, drikke- og tobaksvareindustri",
        "grp127kode": 12000,
        "grp127hovedafdeling": "Tobaksindustri",
    },
    131000: {
        "branche": "Forbehandling og spinding af tekstilfibre",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 13000,
        "grp127hovedafdeling": "Tekstilindustri",
    },
    132000: {
        "branche": "Vævning af tekstiler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 13000,
        "grp127hovedafdeling": "Tekstilindustri",
    },
    133000: {
        "branche": "Efterbehandling af tekstiler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 13000,
        "grp127hovedafdeling": "Tekstilindustri",
    },
    139100: {
        "branche": "Fremstilling af trikotagestoffer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 13000,
        "grp127hovedafdeling": "Tekstilindustri",
    },
    139210: {
        "branche": "Fremstilling af boligtekstiler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 13000,
        "grp127hovedafdeling": "Tekstilindustri",
    },
    139220: {
        "branche": "Fremstilling af færdige tekstilvarer undtagen boligtekstiler og beklædningsartikler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 13000,
        "grp127hovedafdeling": "Tekstilindustri",
    },
    139300: {
        "branche": "Fremstilling af tæpper",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 13000,
        "grp127hovedafdeling": "Tekstilindustri",
    },
    139400: {
        "branche": "Fremstilling af reb, tovværk, sejlgarn og netstoffer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 13000,
        "grp127hovedafdeling": "Tekstilindustri",
    },
    139500: {
        "branche": "Fremstilling af fiberdug og varer af fiberdug undtagen beklædningsartikler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 13000,
        "grp127hovedafdeling": "Tekstilindustri",
    },
    139600: {
        "branche": "Fremstilling af andre tekniske og industrielle tekstiler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 13000,
        "grp127hovedafdeling": "Tekstilindustri",
    },
    139900: {
        "branche": "Fremstilling af andre tekstiler i.a.n.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 13000,
        "grp127hovedafdeling": "Tekstilindustri",
    },
    141100: {
        "branche": "Fremstilling af beklædningsartikler af læder",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 14000,
        "grp127hovedafdeling": "Beklædningsindustri",
    },
    141200: {
        "branche": "Fremstilling af arbejdsbeklædning",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 14000,
        "grp127hovedafdeling": "Beklædningsindustri",
    },
    141300: {
        "branche": "Fremstilling af anden yderbeklædning",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 14000,
        "grp127hovedafdeling": "Beklædningsindustri",
    },
    141400: {
        "branche": "Fremstilling af underbeklædning",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 14000,
        "grp127hovedafdeling": "Beklædningsindustri",
    },
    141900: {
        "branche": "Fremstilling af andre beklædningsartikler samt tilbehør",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 14000,
        "grp127hovedafdeling": "Beklædningsindustri",
    },
    142000: {
        "branche": "Fremstilling af varer af pelsskind",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 14000,
        "grp127hovedafdeling": "Beklædningsindustri",
    },
    143100: {
        "branche": "Fremstilling af strikkede og hæklede strømpevarer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 14000,
        "grp127hovedafdeling": "Beklædningsindustri",
    },
    143900: {
        "branche": "Fremstilling af andre strikkede og hæklede beklædningsartikler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 14000,
        "grp127hovedafdeling": "Beklædningsindustri",
    },
    151100: {
        "branche": "Garvning og beredning af læder; beredning og farvning af pelsskind",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 15000,
        "grp127hovedafdeling": "Læder- og fodtøjsindustri",
    },
    151200: {
        "branche": "Fremstilling af tasker, kufferter, sadelmagervarer mv.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 15000,
        "grp127hovedafdeling": "Læder- og fodtøjsindustri",
    },
    152000: {
        "branche": "Fremstilling af fodtøj",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CB",
        "grp36hovedafdeling": "Tekstil- og læderindustri",
        "grp127kode": 15000,
        "grp127hovedafdeling": "Læder- og fodtøjsindustri",
    },
    161000: {
        "branche": "Udsavning og høvling af træ",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 16000,
        "grp127hovedafdeling": "Træindustri",
    },
    162100: {
        "branche": "Fremstilling af finerplader og træbaserede plader",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 16000,
        "grp127hovedafdeling": "Træindustri",
    },
    162200: {
        "branche": "Fremstilling af sammensatte parketstave",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 16000,
        "grp127hovedafdeling": "Træindustri",
    },
    162300: {
        "branche": "Fremstilling af bygningstømmer og snedkeriartikler i øvrigt",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 16000,
        "grp127hovedafdeling": "Træindustri",
    },
    162400: {
        "branche": "Fremstilling af træemballage",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 16000,
        "grp127hovedafdeling": "Træindustri",
    },
    162900: {
        "branche": "Fremstilling af andre træprodukter; fremstilling af varer af kork, strå og flettematerialer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 16000,
        "grp127hovedafdeling": "Træindustri",
    },
    171100: {
        "branche": "Fremstilling af papirmasse",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 17000,
        "grp127hovedafdeling": "Papirindustri",
    },
    171200: {
        "branche": "Fremstilling af papir og pap",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 17000,
        "grp127hovedafdeling": "Papirindustri",
    },
    172100: {
        "branche": "Fremstilling af bølgepap og pap og emballage af papir og pap",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 17000,
        "grp127hovedafdeling": "Papirindustri",
    },
    172200: {
        "branche": "Fremstilling af husholdningsartikler og hygiejneartikler samt toiletartikler af papir og pap",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 17000,
        "grp127hovedafdeling": "Papirindustri",
    },
    172300: {
        "branche": "Fremstilling af kontorartikler af papir",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 17000,
        "grp127hovedafdeling": "Papirindustri",
    },
    172400: {
        "branche": "Fremstilling af tapet",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 17000,
        "grp127hovedafdeling": "Papirindustri",
    },
    172900: {
        "branche": "Fremstilling af andre papir- og papvarer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 17000,
        "grp127hovedafdeling": "Papirindustri",
    },
    181100: {
        "branche": "Trykning af dagblade",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 18000,
        "grp127hovedafdeling": "Trykkerier mv.",
    },
    181200: {
        "branche": "Anden trykning",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 18000,
        "grp127hovedafdeling": "Trykkerier mv.",
    },
    181300: {
        "branche": "Prepress- og premedia-arbejde",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 18000,
        "grp127hovedafdeling": "Trykkerier mv.",
    },
    181400: {
        "branche": "Bogbinding og lignende serviceydelser",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 18000,
        "grp127hovedafdeling": "Trykkerier mv.",
    },
    182000: {
        "branche": "Reproduktion af indspillede medier",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CC",
        "grp36hovedafdeling": "Træ- og papirindustri, trykkerier",
        "grp127kode": 18000,
        "grp127hovedafdeling": "Trykkerier mv.",
    },
    191000: {
        "branche": "Fremstilling af koks mv.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CD",
        "grp36hovedafdeling": "Olieraffinaderier mv.",
        "grp127kode": 19000,
        "grp127hovedafdeling": "Olieraffinaderier mv. - Olieraffinaderier mv. ",
    },
    192000: {
        "branche": "Fremstilling af raffinerede mineralolieprodukter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CD",
        "grp36hovedafdeling": "Olieraffinaderier mv.",
        "grp127kode": 19000,
        "grp127hovedafdeling": "Olieraffinaderier mv. - Olieraffinaderier mv. ",
    },
    201100: {
        "branche": "Fremstilling af industrigasser",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20001,
        "grp127hovedafdeling": "Fremst. af basiskemikalier",
    },
    201200: {
        "branche": "Fremstilling af farvestoffer og pigmenter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20001,
        "grp127hovedafdeling": "Fremst. af basiskemikalier",
    },
    201300: {
        "branche": "Fremstilling af andre uorganiske basiskemikalier",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20001,
        "grp127hovedafdeling": "Fremst. af basiskemikalier",
    },
    201400: {
        "branche": "Fremstilling af andre organiske basiskemikalier",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20001,
        "grp127hovedafdeling": "Fremst. af basiskemikalier",
    },
    201500: {
        "branche": "Fremstilling af gødningsstoffer og nitrogenprodukter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20001,
        "grp127hovedafdeling": "Fremst. af basiskemikalier",
    },
    201600: {
        "branche": "Fremstilling af plast i ubearbejdet form",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20001,
        "grp127hovedafdeling": "Fremst. af basiskemikalier",
    },
    201700: {
        "branche": "Fremstilling af syntetisk gummi i ubearbejdet form",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20001,
        "grp127hovedafdeling": "Fremst. af basiskemikalier",
    },
    202000: {
        "branche": "Fremstilling af pesticider og andre agrokemiske produkter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20001,
        "grp127hovedafdeling": "Fremst. af basiskemikalier",
    },
    203000: {
        "branche": "Fremstilling af maling, lak og lignende overfladebehandlingsmidler, trykfarver samt tætningsmaterialer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20002,
        "grp127hovedafdeling": "Fremst. af maling og sæbe mv.",
    },
    204100: {
        "branche": "Fremstilling af sæbe, rengørings- og rensemidler samt poleremidler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20002,
        "grp127hovedafdeling": "Fremst. af maling og sæbe mv.",
    },
    204200: {
        "branche": "Fremstilling af parfume, hårshampoo, tandpasta mv.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20002,
        "grp127hovedafdeling": "Fremst. af maling og sæbe mv.",
    },
    205100: {
        "branche": "Fremstilling af sprængstoffer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20002,
        "grp127hovedafdeling": "Fremst. af maling og sæbe mv.",
    },
    205200: {
        "branche": "Fremstilling af lim",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20002,
        "grp127hovedafdeling": "Fremst. af maling og sæbe mv.",
    },
    205300: {
        "branche": "Fremstilling af æteriske olier",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20002,
        "grp127hovedafdeling": "Fremst. af maling og sæbe mv.",
    },
    205900: {
        "branche": "Fremstilling af andre kemiske produkter i.a.n.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20002,
        "grp127hovedafdeling": "Fremst. af maling og sæbe mv.",
    },
    206000: {
        "branche": "Fremstilling af kemofibre",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CE",
        "grp36hovedafdeling": "Kemisk industri",
        "grp127kode": 20002,
        "grp127hovedafdeling": "Fremst. af maling og sæbe mv.",
    },
    211000: {
        "branche": "Fremstilling af farmaceutiske råvarer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CF",
        "grp36hovedafdeling": "Medicinalindustri",
        "grp127kode": 21000,
        "grp127hovedafdeling": "Medicinalindustri",
    },
    212000: {
        "branche": "Fremstilling af farmaceutiske præparater",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CF",
        "grp36hovedafdeling": "Medicinalindustri",
        "grp127kode": 21000,
        "grp127hovedafdeling": "Medicinalindustri",
    },
    221100: {
        "branche": "Fremstilling af gummidæk og gummislanger; vulkanisering af dæk",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 22000,
        "grp127hovedafdeling": "Plast- og gummiindustri",
    },
    221900: {
        "branche": "Fremstilling af andre gummiprodukter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 22000,
        "grp127hovedafdeling": "Plast- og gummiindustri",
    },
    222100: {
        "branche": "Fremstilling af plader, ark, rør og slanger samt profiler af plast",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 22000,
        "grp127hovedafdeling": "Plast- og gummiindustri",
    },
    222200: {
        "branche": "Fremstilling af plastemballage",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 22000,
        "grp127hovedafdeling": "Plast- og gummiindustri",
    },
    222300: {
        "branche": "Fremstilling af bygningsartikler af plast",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 22000,
        "grp127hovedafdeling": "Plast- og gummiindustri",
    },
    222900: {
        "branche": "Fremstilling af andre plastprodukter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 22000,
        "grp127hovedafdeling": "Plast- og gummiindustri",
    },
    231100: {
        "branche": "Fremstilling af planglas",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23001,
        "grp127hovedafdeling": "Glasindustri og keramisk industri",
    },
    231200: {
        "branche": "Formning og forarbejdning af planglas",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23001,
        "grp127hovedafdeling": "Glasindustri og keramisk industri",
    },
    231300: {
        "branche": "Fremstilling af flasker, drikkeglas mv.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23001,
        "grp127hovedafdeling": "Glasindustri og keramisk industri",
    },
    231400: {
        "branche": "Fremstilling af glasfiber",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23001,
        "grp127hovedafdeling": "Glasindustri og keramisk industri",
    },
    231900: {
        "branche": "Fremstilling og bearbejdning af andet glas (herunder teknisk glas)",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23001,
        "grp127hovedafdeling": "Glasindustri og keramisk industri",
    },
    232000: {
        "branche": "Fremstilling af ildfaste produkter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23001,
        "grp127hovedafdeling": "Glasindustri og keramisk industri",
    },
    233100: {
        "branche": "Fremstilling af keramiske teglsten og gulvfliser",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23002,
        "grp127hovedafdeling": "Betonindustri og teglværker",
    },
    233200: {
        "branche": "Fremstilling af mursten, teglsten og byggematerialer af brændt ler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23002,
        "grp127hovedafdeling": "Betonindustri og teglværker",
    },
    234100: {
        "branche": "Fremstilling af keramiske husholdningsartikler og pyntegenstande",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23001,
        "grp127hovedafdeling": "Glasindustri og keramisk industri",
    },
    234200: {
        "branche": "Fremstilling af keramiske sanitetsartikler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23001,
        "grp127hovedafdeling": "Glasindustri og keramisk industri",
    },
    234300: {
        "branche": "Fremstilling af keramiske isolatorer og isoleringsdele",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23001,
        "grp127hovedafdeling": "Glasindustri og keramisk industri",
    },
    234400: {
        "branche": "Fremstilling af andre keramiske produkter til teknisk brug",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23001,
        "grp127hovedafdeling": "Glasindustri og keramisk industri",
    },
    234900: {
        "branche": "Fremstilling af andre keramiske produkter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23001,
        "grp127hovedafdeling": "Glasindustri og keramisk industri",
    },
    235100: {
        "branche": "Fremstilling af cement",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23002,
        "grp127hovedafdeling": "Betonindustri og teglværker",
    },
    235200: {
        "branche": "Fremstilling af kalk og gips",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23002,
        "grp127hovedafdeling": "Betonindustri og teglværker",
    },
    236100: {
        "branche": "Fremstilling af byggematerialer af beton",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23002,
        "grp127hovedafdeling": "Betonindustri og teglværker",
    },
    236200: {
        "branche": "Fremstilling af byggematerialer af gips",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23002,
        "grp127hovedafdeling": "Betonindustri og teglværker",
    },
    236300: {
        "branche": "Fremstilling af færdigblandet beton",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23002,
        "grp127hovedafdeling": "Betonindustri og teglværker",
    },
    236400: {
        "branche": "Fremstilling af mørtel",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23002,
        "grp127hovedafdeling": "Betonindustri og teglværker",
    },
    236500: {
        "branche": "Fremstilling af fibercement",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23002,
        "grp127hovedafdeling": "Betonindustri og teglværker",
    },
    236900: {
        "branche": "Fremstilling af andre beton-, gips- og cementprodukter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23002,
        "grp127hovedafdeling": "Betonindustri og teglværker",
    },
    237000: {
        "branche": "Tilhugning, tilskæring og færdigbearbejdning af sten",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23002,
        "grp127hovedafdeling": "Betonindustri og teglværker",
    },
    239100: {
        "branche": "Fremstilling af slibemidler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23002,
        "grp127hovedafdeling": "Betonindustri og teglværker",
    },
    239910: {
        "branche": "Fremstilling af asfalt og tagpap",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23002,
        "grp127hovedafdeling": "Betonindustri og teglværker",
    },
    239990: {
        "branche": "Fremstilling af andre ikke-metalholdige mineralske produkter i.a.n.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CG",
        "grp36hovedafdeling": "Plast-, glas- og betonindustri",
        "grp127kode": 23002,
        "grp127hovedafdeling": "Betonindustri og teglværker",
    },
    241000: {
        "branche": "Fremstilling af råjern og råstål samt jernlegeringer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    242000: {
        "branche": "Fremstilling af rør og hule profiler og tilhørende fittings af stål",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    243100: {
        "branche": "Fremstilling af stænger ved koldtrækning",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    243200: {
        "branche": "Fremstilling af stålbånd ved koldvalsning",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    243300: {
        "branche": "Koldbehandling",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    243400: {
        "branche": "Fremstilling af tråd ved koldtrækning",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    244100: {
        "branche": "Fremstilling af ædelmetaller",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    244200: {
        "branche": "Fremstilling af aluminium",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    244300: {
        "branche": "Fremstilling af bly, zink og tin",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    244400: {
        "branche": "Fremstilling af kobber",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    244500: {
        "branche": "Anden fremstilling af ikke-jernholdige metaller",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    244600: {
        "branche": "Oparbejdning af nukleart brændsel",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    245100: {
        "branche": "Støbning af jernprodukter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    245200: {
        "branche": "Støbning af stålprodukter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    245300: {
        "branche": "Støbning af letmetalprodukter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    245400: {
        "branche": "Støbning af andre ikke-jernholdige metalprodukter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 24000,
        "grp127hovedafdeling": "Fremst. af metal",
    },
    251100: {
        "branche": "Fremstilling af metalkonstruktioner og dele heraf",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    251200: {
        "branche": "Fremstilling af døre og vinduer af metal",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    252100: {
        "branche": "Fremstilling af radiatorer og kedler til centralvarmeanlæg",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    252900: {
        "branche": "Fremstilling af andre tanke og beholdere af metal",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    253000: {
        "branche": "Fremstilling af dampkedler undtagen centralvarmekedler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    254000: {
        "branche": "Fremstilling af våben og ammunition",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    255000: {
        "branche": "Smedning, presning, sænksmedning og valsning af metal; pulvermetallurgi",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    256100: {
        "branche": "Overfladebehandling af metal",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    256200: {
        "branche": "Maskinforarbejdning",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    257100: {
        "branche": "Fremstilling af bestik, skære- og klipperedskaber",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    257200: {
        "branche": "Fremstilling af låse og hængsler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    257300: {
        "branche": "Fremstilling af håndværktøj",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    259100: {
        "branche": "Fremstilling af metaltønder og lignende beholdere",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    259200: {
        "branche": "Fremstilling af letmetalemballage",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    259300: {
        "branche": "Fremstilling af trådvarer, kæder og fjedre",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    259400: {
        "branche": "Fremstilling af lukkeanordninger, bolte, skruer og møtrikker",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    259900: {
        "branche": "Fremstilling af andre færdige metalprodukter i.a.n.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CH",
        "grp36hovedafdeling": "Metalindustri",
        "grp127kode": 25000,
        "grp127hovedafdeling": "Metalvareindustri",
    },
    261100: {
        "branche": "Fremstilling af elektroniske komponenter og plader",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CI",
        "grp36hovedafdeling": "Elektronikindustri",
        "grp127kode": 26001,
        "grp127hovedafdeling": "Fremst. af computere og kommunikationsudstyr mv. - Fremst. af computere og kommunikationsudstyr mv. ",
    },
    261200: {
        "branche": "Fremstilling af printplader o.l.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CI",
        "grp36hovedafdeling": "Elektronikindustri",
        "grp127kode": 26001,
        "grp127hovedafdeling": "Fremst. af computere og kommunikationsudstyr mv. - Fremst. af computere og kommunikationsudstyr mv. ",
    },
    262000: {
        "branche": "Fremstilling af computere og ydre enheder",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CI",
        "grp36hovedafdeling": "Elektronikindustri",
        "grp127kode": 26001,
        "grp127hovedafdeling": "Fremst. af computere og kommunikationsudstyr mv. - Fremst. af computere og kommunikationsudstyr mv. ",
    },
    263000: {
        "branche": "Fremstilling af kommunikationsudstyr",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CI",
        "grp36hovedafdeling": "Elektronikindustri",
        "grp127kode": 26001,
        "grp127hovedafdeling": "Fremst. af computere og kommunikationsudstyr mv. - Fremst. af computere og kommunikationsudstyr mv. ",
    },
    264000: {
        "branche": "Fremstilling af elektronik til husholdninger",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CI",
        "grp36hovedafdeling": "Elektronikindustri",
        "grp127kode": 26001,
        "grp127hovedafdeling": "Fremst. af computere og kommunikationsudstyr mv. - Fremst. af computere og kommunikationsudstyr mv. ",
    },
    265100: {
        "branche": "Fremstilling af udstyr til måling, afprøvning, navigation og kontrol",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CI",
        "grp36hovedafdeling": "Elektronikindustri",
        "grp127kode": 26002,
        "grp127hovedafdeling": "Fremst. af andet elektronisk udstyr",
    },
    265200: {
        "branche": "Fremstilling af ure",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CI",
        "grp36hovedafdeling": "Elektronikindustri",
        "grp127kode": 26002,
        "grp127hovedafdeling": "Fremst. af andet elektronisk udstyr",
    },
    266010: {
        "branche": "Fremstilling af høreapparater og dele hertil",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CI",
        "grp36hovedafdeling": "Elektronikindustri",
        "grp127kode": 26002,
        "grp127hovedafdeling": "Fremst. af andet elektronisk udstyr",
    },
    266090: {
        "branche": "Fremstilling af bestrålingsudstyr og elektromedicinsk og elektroterapeutisk udstyr",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CI",
        "grp36hovedafdeling": "Elektronikindustri",
        "grp127kode": 26002,
        "grp127hovedafdeling": "Fremst. af andet elektronisk udstyr",
    },
    267000: {
        "branche": "Fremstilling af optiske instrumenter og fotografisk udstyr",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CI",
        "grp36hovedafdeling": "Elektronikindustri",
        "grp127kode": 26002,
        "grp127hovedafdeling": "Fremst. af andet elektronisk udstyr",
    },
    268000: {
        "branche": "Fremstilling af magnetiske og optiske media",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CI",
        "grp36hovedafdeling": "Elektronikindustri",
        "grp127kode": 26002,
        "grp127hovedafdeling": "Fremst. af andet elektronisk udstyr",
    },
    271100: {
        "branche": "Fremstilling af elektriske motorer, generatorer og transformere",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CJ",
        "grp36hovedafdeling": "Fremst. af elektrisk udstyr",
        "grp127kode": 27001,
        "grp127hovedafdeling": "Fremst. af elektriske motorer mv. - Fremst. af elektriske motorer mv. ",
    },
    271200: {
        "branche": "Fremstilling af elektriske fordelings- og kontrolapparater",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CJ",
        "grp36hovedafdeling": "Fremst. af elektrisk udstyr",
        "grp127kode": 27001,
        "grp127hovedafdeling": "Fremst. af elektriske motorer mv. - Fremst. af elektriske motorer mv. ",
    },
    272000: {
        "branche": "Fremstilling af batterier og akkumulatorer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CJ",
        "grp36hovedafdeling": "Fremst. af elektrisk udstyr",
        "grp127kode": 27001,
        "grp127hovedafdeling": "Fremst. af elektriske motorer mv. - Fremst. af elektriske motorer mv. ",
    },
    273100: {
        "branche": "Fremstilling af lyslederkabler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CJ",
        "grp36hovedafdeling": "Fremst. af elektrisk udstyr",
        "grp127kode": 27002,
        "grp127hovedafdeling": "Fremst. af ledninger og kabler",
    },
    273200: {
        "branche": "Fremstilling af andre elektroniske og elektriske ledninger og kabler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CJ",
        "grp36hovedafdeling": "Fremst. af elektrisk udstyr",
        "grp127kode": 27002,
        "grp127hovedafdeling": "Fremst. af ledninger og kabler",
    },
    273300: {
        "branche": "Fremstilling af tilbehør til ledninger og kabler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CJ",
        "grp36hovedafdeling": "Fremst. af elektrisk udstyr",
        "grp127kode": 27002,
        "grp127hovedafdeling": "Fremst. af ledninger og kabler",
    },
    274000: {
        "branche": "Fremstilling af elektriske belysningsartikler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CJ",
        "grp36hovedafdeling": "Fremst. af elektrisk udstyr",
        "grp127kode": 27003,
        "grp127hovedafdeling": "Fremst. af husholdningsapparater, lamper mv.",
    },
    275100: {
        "branche": "Fremstilling af elektriske husholdningsapparater",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CJ",
        "grp36hovedafdeling": "Fremst. af elektrisk udstyr",
        "grp127kode": 27003,
        "grp127hovedafdeling": "Fremst. af husholdningsapparater, lamper mv.",
    },
    275200: {
        "branche": "Fremstilling af ikke-elektriske husholdningsapparater",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CJ",
        "grp36hovedafdeling": "Fremst. af elektrisk udstyr",
        "grp127kode": 27003,
        "grp127hovedafdeling": "Fremst. af husholdningsapparater, lamper mv.",
    },
    279000: {
        "branche": "Fremstilling af andet elektrisk udstyr",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CJ",
        "grp36hovedafdeling": "Fremst. af elektrisk udstyr",
        "grp127kode": 27003,
        "grp127hovedafdeling": "Fremst. af husholdningsapparater, lamper mv.",
    },
    281110: {
        "branche": "Fremstilling af vindmøller og dele hertil",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28001,
        "grp127hovedafdeling": "Fremst. af motorer, vindmøller og pumper",
    },
    281190: {
        "branche": "Fremstilling af motorer og turbiner undtagen motorer til vindmøller, flyvemaskiner, motorkøretøjer og knallerter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28001,
        "grp127hovedafdeling": "Fremst. af motorer, vindmøller og pumper",
    },
    281200: {
        "branche": "Fremstilling af hydraulisk udstyr",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28001,
        "grp127hovedafdeling": "Fremst. af motorer, vindmøller og pumper",
    },
    281300: {
        "branche": "Fremstilling af andre pumper og kompressorer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28001,
        "grp127hovedafdeling": "Fremst. af motorer, vindmøller og pumper",
    },
    281400: {
        "branche": "Fremstilling af andre haner og ventiler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28001,
        "grp127hovedafdeling": "Fremst. af motorer, vindmøller og pumper",
    },
    281500: {
        "branche": "Fremstilling af lejer, tandhjul, tandhjulsudvekslinger og drivelementer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28001,
        "grp127hovedafdeling": "Fremst. af motorer, vindmøller og pumper",
    },
    282100: {
        "branche": "Fremstilling af ovne, ildsteder og fyringsaggregater",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    282200: {
        "branche": "Fremstilling af løfte- og håndteringsudstyr",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    282300: {
        "branche": "Fremstilling af kontormaskiner og -udstyr (undtagen computere og ydre enheder)",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    282400: {
        "branche": "Fremstilling af motordrevet håndværktøj",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    282500: {
        "branche": "Fremstilling af køle- og ventilationsanlæg (til industriel brug)",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    282900: {
        "branche": "Fremstilling af andre maskiner til generelle formål i.a.n.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    283000: {
        "branche": "Fremstilling af landbrugs- og skovbrugsmaskiner",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    284100: {
        "branche": "Fremstilling af metalforarbejdende værktøjsmaskiner",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    284900: {
        "branche": "Fremstilling af andre værktøjsmaskiner",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    289100: {
        "branche": "Fremstilling af maskiner til metallurgi",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    289200: {
        "branche": "Fremstilling af maskiner til råstofindvindingsindustrien samt bygge og anlæg",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    289300: {
        "branche": "Fremstilling af maskiner til føde-, drikke- og tobaksvareindustrien",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    289400: {
        "branche": "Fremstilling af maskiner til produktion af tekstiler, beklædningsartikler og læder",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    289500: {
        "branche": "Fremstilling af maskiner til produktion af papir og pap",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    289600: {
        "branche": "Fremstilling af maskiner til produktion af plast og gummi",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    289900: {
        "branche": "Fremstilling af øvrige maskiner til specielle formål i.a.n.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CK",
        "grp36hovedafdeling": "Maskinindustri - Maskinindustri ",
        "grp127kode": 28002,
        "grp127hovedafdeling": "Fremst. af andre maskiner",
    },
    291000: {
        "branche": "Fremstilling af motorkøretøjer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CL",
        "grp36hovedafdeling": "Transportmiddelindustri",
        "grp127kode": 29000,
        "grp127hovedafdeling": "Fremst. af motorkøretøjer og dele hertil",
    },
    292000: {
        "branche": "Fremstilling af karosserier til motorkøretøjer; fremstilling af påhængsvogne og sættevogne",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CL",
        "grp36hovedafdeling": "Transportmiddelindustri",
        "grp127kode": 29000,
        "grp127hovedafdeling": "Fremst. af motorkøretøjer og dele hertil",
    },
    293100: {
        "branche": "Fremstilling af elektrisk og elektronisk udstyr til motorkøretøjer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CL",
        "grp36hovedafdeling": "Transportmiddelindustri",
        "grp127kode": 29000,
        "grp127hovedafdeling": "Fremst. af motorkøretøjer og dele hertil",
    },
    293200: {
        "branche": "Fremstilling af andre dele og tilbehør til motorkøretøjer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CL",
        "grp36hovedafdeling": "Transportmiddelindustri",
        "grp127kode": 29000,
        "grp127hovedafdeling": "Fremst. af motorkøretøjer og dele hertil",
    },
    301100: {
        "branche": "Bygning af skibe og flydende materiel",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CL",
        "grp36hovedafdeling": "Transportmiddelindustri",
        "grp127kode": 30000,
        "grp127hovedafdeling": "Fremst. af skibe og andre transportmidler",
    },
    301200: {
        "branche": "Bygning af både til fritid og sport",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CL",
        "grp36hovedafdeling": "Transportmiddelindustri",
        "grp127kode": 30000,
        "grp127hovedafdeling": "Fremst. af skibe og andre transportmidler",
    },
    302000: {
        "branche": "Fremstilling af lokomotiver og andet rullende materiel til jernbaner og sporveje",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CL",
        "grp36hovedafdeling": "Transportmiddelindustri",
        "grp127kode": 30000,
        "grp127hovedafdeling": "Fremst. af skibe og andre transportmidler",
    },
    303000: {
        "branche": "Fremstilling af luft- og rumfartøjer o.l.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CL",
        "grp36hovedafdeling": "Transportmiddelindustri",
        "grp127kode": 30000,
        "grp127hovedafdeling": "Fremst. af skibe og andre transportmidler",
    },
    304000: {
        "branche": "Fremstilling af militære kampkøretøjer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CL",
        "grp36hovedafdeling": "Transportmiddelindustri",
        "grp127kode": 30000,
        "grp127hovedafdeling": "Fremst. af skibe og andre transportmidler",
    },
    309100: {
        "branche": "Fremstilling af motorcykler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CL",
        "grp36hovedafdeling": "Transportmiddelindustri",
        "grp127kode": 30000,
        "grp127hovedafdeling": "Fremst. af skibe og andre transportmidler",
    },
    309200: {
        "branche": "Fremstilling af cykler og invalidekøretøjer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CL",
        "grp36hovedafdeling": "Transportmiddelindustri",
        "grp127kode": 30000,
        "grp127hovedafdeling": "Fremst. af skibe og andre transportmidler",
    },
    309900: {
        "branche": "Fremstilling af andre transportmidler i.a.n.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CL",
        "grp36hovedafdeling": "Transportmiddelindustri",
        "grp127kode": 30000,
        "grp127hovedafdeling": "Fremst. af skibe og andre transportmidler",
    },
    310100: {
        "branche": "Fremstilling af kontor- og butiksmøbler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 31000,
        "grp127hovedafdeling": "Møbelindustri",
    },
    310200: {
        "branche": "Fremstilling af køkkenmøbler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 31000,
        "grp127hovedafdeling": "Møbelindustri",
    },
    310300: {
        "branche": "Fremstilling af madrasser",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 31000,
        "grp127hovedafdeling": "Møbelindustri",
    },
    310900: {
        "branche": "Fremstilling af andre møbler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 31000,
        "grp127hovedafdeling": "Møbelindustri",
    },
    321100: {
        "branche": "Prægning af mønter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 32002,
        "grp127hovedafdeling": "Legetøj og anden fremstillingsvirksomhed",
    },
    321200: {
        "branche": "Fremstilling af smykker i ædle metaller og relaterede produkter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 32002,
        "grp127hovedafdeling": "Legetøj og anden fremstillingsvirksomhed",
    },
    321300: {
        "branche": "Fremstilling af bijouteri og lignende varer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 32002,
        "grp127hovedafdeling": "Legetøj og anden fremstillingsvirksomhed",
    },
    322000: {
        "branche": "Fremstilling af musikinstrumenter",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 32002,
        "grp127hovedafdeling": "Legetøj og anden fremstillingsvirksomhed",
    },
    323000: {
        "branche": "Fremstilling af sportsudstyr",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 32002,
        "grp127hovedafdeling": "Legetøj og anden fremstillingsvirksomhed",
    },
    324000: {
        "branche": "Fremstilling af spil og legetøj",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 32002,
        "grp127hovedafdeling": "Legetøj og anden fremstillingsvirksomhed",
    },
    325000: {
        "branche": "Fremstilling af medicinske og dentale instrumenter samt udstyr hertil",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 32002,
        "grp127hovedafdeling": "Legetøj og anden fremstillingsvirksomhed",
    },
    329100: {
        "branche": "Fremstilling af koste og børster",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 32002,
        "grp127hovedafdeling": "Legetøj og anden fremstillingsvirksomhed",
    },
    329900: {
        "branche": "Anden fremstillingsvirksomhed i.a.n.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 32002,
        "grp127hovedafdeling": "Legetøj og anden fremstillingsvirksomhed",
    },
    331100: {
        "branche": "Reparation af jern- og metalvarer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 33000,
        "grp127hovedafdeling": "Reparation og installation af maskiner og udstyr",
    },
    331200: {
        "branche": "Reparation af maskiner",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 33000,
        "grp127hovedafdeling": "Reparation og installation af maskiner og udstyr",
    },
    331300: {
        "branche": "Reparation af elektronisk og optisk udstyr",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 33000,
        "grp127hovedafdeling": "Reparation og installation af maskiner og udstyr",
    },
    331400: {
        "branche": "Reparation af elektrisk udstyr",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 33000,
        "grp127hovedafdeling": "Reparation og installation af maskiner og udstyr",
    },
    331500: {
        "branche": "Reparation og vedligeholdelse af skibe og både",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 33000,
        "grp127hovedafdeling": "Reparation og installation af maskiner og udstyr",
    },
    331600: {
        "branche": "Reparation og vedligeholdelse af luft- og rumfartøjer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 33000,
        "grp127hovedafdeling": "Reparation og installation af maskiner og udstyr",
    },
    331700: {
        "branche": "Reparation og vedligeholdelse af andre transportmidler",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 33000,
        "grp127hovedafdeling": "Reparation og installation af maskiner og udstyr",
    },
    331900: {
        "branche": "Reparation af andet udstyr",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 33000,
        "grp127hovedafdeling": "Reparation og installation af maskiner og udstyr",
    },
    332000: {
        "branche": "Installation af industrimaskiner og -udstyr",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "C",
        "grp19hovedafdeling": "Industri - Industri ",
        "grp36kode": "CM",
        "grp36hovedafdeling": "Møbel og anden industri mv.",
        "grp127kode": 33000,
        "grp127hovedafdeling": "Reparation og installation af maskiner og udstyr",
    },
    351100: {
        "branche": "Produktion af elektricitet",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "D",
        "grp19hovedafdeling": "Energiforsyning",
        "grp36kode": "D",
        "grp36hovedafdeling": "Energiforsyning - Energiforsyning ",
        "grp127kode": 35001,
        "grp127hovedafdeling": "Elforsyning",
    },
    351200: {
        "branche": "Transmission af elektricitet",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "D",
        "grp19hovedafdeling": "Energiforsyning",
        "grp36kode": "D",
        "grp36hovedafdeling": "Energiforsyning - Energiforsyning ",
        "grp127kode": 35001,
        "grp127hovedafdeling": "Elforsyning",
    },
    351300: {
        "branche": "Distribution af elektricitet",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "D",
        "grp19hovedafdeling": "Energiforsyning",
        "grp36kode": "D",
        "grp36hovedafdeling": "Energiforsyning - Energiforsyning ",
        "grp127kode": 35001,
        "grp127hovedafdeling": "Elforsyning",
    },
    351400: {
        "branche": "Handel med elektricitet",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "D",
        "grp19hovedafdeling": "Energiforsyning",
        "grp36kode": "D",
        "grp36hovedafdeling": "Energiforsyning - Energiforsyning ",
        "grp127kode": 35001,
        "grp127hovedafdeling": "Elforsyning",
    },
    352100: {
        "branche": "Fremstilling af gas",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "D",
        "grp19hovedafdeling": "Energiforsyning",
        "grp36kode": "D",
        "grp36hovedafdeling": "Energiforsyning - Energiforsyning ",
        "grp127kode": 35002,
        "grp127hovedafdeling": "Gasforsyning",
    },
    352200: {
        "branche": "Distribution af gas",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "D",
        "grp19hovedafdeling": "Energiforsyning",
        "grp36kode": "D",
        "grp36hovedafdeling": "Energiforsyning - Energiforsyning ",
        "grp127kode": 35002,
        "grp127hovedafdeling": "Gasforsyning",
    },
    352300: {
        "branche": "Handel med gas gennem rørledninger",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "D",
        "grp19hovedafdeling": "Energiforsyning",
        "grp36kode": "D",
        "grp36hovedafdeling": "Energiforsyning - Energiforsyning ",
        "grp127kode": 35002,
        "grp127hovedafdeling": "Gasforsyning",
    },
    353000: {
        "branche": "Varmeforsyning",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "D",
        "grp19hovedafdeling": "Energiforsyning",
        "grp36kode": "D",
        "grp36hovedafdeling": "Energiforsyning - Energiforsyning ",
        "grp127kode": 35003,
        "grp127hovedafdeling": "Varmeforsyning",
    },
    360000: {
        "branche": "Vandforsyning",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "E",
        "grp19hovedafdeling": "Vandforsyning og renovation",
        "grp36kode": "E",
        "grp36hovedafdeling": "Vandforsyning og renovation - Vandforsyning og renovation  ",
        "grp127kode": 36000,
        "grp127hovedafdeling": "Vandforsyning",
    },
    370000: {
        "branche": "Opsamling og behandling af spildevand",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "E",
        "grp19hovedafdeling": "Vandforsyning og renovation",
        "grp36kode": "E",
        "grp36hovedafdeling": "Vandforsyning og renovation - Vandforsyning og renovation  ",
        "grp127kode": 37000,
        "grp127hovedafdeling": "Kloak- og rensningsanlæg",
    },
    381100: {
        "branche": "Indsamling af ikke-farligt affald",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "E",
        "grp19hovedafdeling": "Vandforsyning og renovation",
        "grp36kode": "E",
        "grp36hovedafdeling": "Vandforsyning og renovation - Vandforsyning og renovation  ",
        "grp127kode": 38000,
        "grp127hovedafdeling": "Renovation og genbrug",
    },
    381200: {
        "branche": "Indsamling af farligt affald",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "E",
        "grp19hovedafdeling": "Vandforsyning og renovation",
        "grp36kode": "E",
        "grp36hovedafdeling": "Vandforsyning og renovation - Vandforsyning og renovation  ",
        "grp127kode": 38000,
        "grp127hovedafdeling": "Renovation og genbrug",
    },
    382110: {
        "branche": "Behandling og bortskaffelse af ikke-farligt affald",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "E",
        "grp19hovedafdeling": "Vandforsyning og renovation",
        "grp36kode": "E",
        "grp36hovedafdeling": "Vandforsyning og renovation - Vandforsyning og renovation  ",
        "grp127kode": 38000,
        "grp127hovedafdeling": "Renovation og genbrug",
    },
    382120: {
        "branche": "Bortskaffelse af affald med energiproduktion",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "E",
        "grp19hovedafdeling": "Vandforsyning og renovation",
        "grp36kode": "E",
        "grp36hovedafdeling": "Vandforsyning og renovation - Vandforsyning og renovation  ",
        "grp127kode": 38000,
        "grp127hovedafdeling": "Renovation og genbrug",
    },
    382200: {
        "branche": "Behandling og bortskaffelse af farligt affald",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "E",
        "grp19hovedafdeling": "Vandforsyning og renovation",
        "grp36kode": "E",
        "grp36hovedafdeling": "Vandforsyning og renovation - Vandforsyning og renovation  ",
        "grp127kode": 38000,
        "grp127hovedafdeling": "Renovation og genbrug",
    },
    383100: {
        "branche": "Demontering af udtjente køretøjer, skibe, maskiner mv.",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "E",
        "grp19hovedafdeling": "Vandforsyning og renovation",
        "grp36kode": "E",
        "grp36hovedafdeling": "Vandforsyning og renovation - Vandforsyning og renovation  ",
        "grp127kode": 38000,
        "grp127hovedafdeling": "Renovation og genbrug",
    },
    383200: {
        "branche": "Genbrug af sorterede materialer",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "E",
        "grp19hovedafdeling": "Vandforsyning og renovation",
        "grp36kode": "E",
        "grp36hovedafdeling": "Vandforsyning og renovation - Vandforsyning og renovation  ",
        "grp127kode": 38000,
        "grp127hovedafdeling": "Renovation og genbrug",
    },
    390000: {
        "branche": "Rensning af jord og grundvand og anden form for forureningsbekæmpelse",
        "grp10kode": 2,
        "grp10hovedafdeling": "Industri, råstofindvinding og forsyningsvirksomhed",
        "grp19kode": "E",
        "grp19hovedafdeling": "Vandforsyning og renovation",
        "grp36kode": "E",
        "grp36hovedafdeling": "Vandforsyning og renovation - Vandforsyning og renovation  ",
        "grp127kode": 39000,
        "grp127hovedafdeling": "Rensning af jord og grundvand",
    },
    411000: {
        "branche": "Gennemførelse af byggeprojekter",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 41000,
        "grp127hovedafdeling": "Byggeentreprenører",
    },
    412000: {
        "branche": "Opførelse af bygninger",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 41000,
        "grp127hovedafdeling": "Byggeentreprenører",
    },
    421100: {
        "branche": "Anlæg af veje og motorveje",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 42000,
        "grp127hovedafdeling": "Anlægsentreprenører",
    },
    421200: {
        "branche": "Anlæg af jernbaner og undergrundsbaner",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 42000,
        "grp127hovedafdeling": "Anlægsentreprenører",
    },
    421300: {
        "branche": "Anlæg af broer og tunneller",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 42000,
        "grp127hovedafdeling": "Anlægsentreprenører",
    },
    422100: {
        "branche": "Anlæg af ledningsnet til væsker",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 42000,
        "grp127hovedafdeling": "Anlægsentreprenører",
    },
    422200: {
        "branche": "Anlæg af ledningsnet til elektricitet og kommunikation",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 42000,
        "grp127hovedafdeling": "Anlægsentreprenører",
    },
    429100: {
        "branche": "Anlæg af vandveje, havne, diger og dæmninger",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 42000,
        "grp127hovedafdeling": "Anlægsentreprenører",
    },
    429900: {
        "branche": "Anden anlægsvirksomhed i.a.n.",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 42000,
        "grp127hovedafdeling": "Anlægsentreprenører",
    },
    431100: {
        "branche": "Nedrivning",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43009,
        "grp127hovedafdeling": "Murere og anden specialiseret bygge- og anlægsvirksomhed samt forberedende byggepladsarbejde - Murere og anden specialiseret bygge- og anlægsvirksomhed samt forberedende byggepladsarbejde ",
    },
    431200: {
        "branche": "Forberedende byggepladsarbejder",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43009,
        "grp127hovedafdeling": "Murere og anden specialiseret bygge- og anlægsvirksomhed samt forberedende byggepladsarbejde - Murere og anden specialiseret bygge- og anlægsvirksomhed samt forberedende byggepladsarbejde ",
    },
    431300: {
        "branche": "Funderingsundersøgelser",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43009,
        "grp127hovedafdeling": "Murere og anden specialiseret bygge- og anlægsvirksomhed samt forberedende byggepladsarbejde - Murere og anden specialiseret bygge- og anlægsvirksomhed samt forberedende byggepladsarbejde ",
    },
    432100: {
        "branche": "El-installation",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43001,
        "grp127hovedafdeling": "Bygningsinstallation",
    },
    432200: {
        "branche": "VVS- og blikkenslagerforretninger",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43001,
        "grp127hovedafdeling": "Bygningsinstallation",
    },
    432900: {
        "branche": "Anden bygningsinstallationsvirksomhed",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43001,
        "grp127hovedafdeling": "Bygningsinstallation",
    },
    433100: {
        "branche": "Stukkatørvirksomhed",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43002,
        "grp127hovedafdeling": "Bygningsfærdiggørelse",
    },
    433200: {
        "branche": "Tømrer- og bygningssnedkervirksomhed",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43002,
        "grp127hovedafdeling": "Bygningsfærdiggørelse",
    },
    433300: {
        "branche": "Udførelse af gulvbelægninger og vægbeklædning",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43002,
        "grp127hovedafdeling": "Bygningsfærdiggørelse",
    },
    433410: {
        "branche": "Malerforretninger",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43002,
        "grp127hovedafdeling": "Bygningsfærdiggørelse",
    },
    433420: {
        "branche": "Glarmestervirksomhed",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43002,
        "grp127hovedafdeling": "Bygningsfærdiggørelse",
    },
    433900: {
        "branche": "Anden bygningsfærdiggørelse",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43002,
        "grp127hovedafdeling": "Bygningsfærdiggørelse",
    },
    439100: {
        "branche": "Tagdækningsvirksomhed",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43009,
        "grp127hovedafdeling": "Murere og anden specialiseret bygge- og anlægsvirksomhed samt forberedende byggepladsarbejde - Murere og anden specialiseret bygge- og anlægsvirksomhed samt forberedende byggepladsarbejde ",
    },
    439910: {
        "branche": "Murere",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43009,
        "grp127hovedafdeling": "Murere og anden specialiseret bygge- og anlægsvirksomhed samt forberedende byggepladsarbejde - Murere og anden specialiseret bygge- og anlægsvirksomhed samt forberedende byggepladsarbejde ",
    },
    439990: {
        "branche": "Anden bygge- og anlægsvirksomhed, som kræver specialisering",
        "grp10kode": 3,
        "grp10hovedafdeling": "Bygge og anlæg",
        "grp19kode": "F",
        "grp19hovedafdeling": "Bygge og anlæg",
        "grp36kode": "F",
        "grp36hovedafdeling": "Bygge og anlæg",
        "grp127kode": 43009,
        "grp127hovedafdeling": "Murere og anden specialiseret bygge- og anlægsvirksomhed samt forberedende byggepladsarbejde - Murere og anden specialiseret bygge- og anlægsvirksomhed samt forberedende byggepladsarbejde ",
    },
    451110: {
        "branche": "Engroshandel med personbiler, varebiler og minibusser",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 45001,
        "grp127hovedafdeling": "Bilhandel",
    },
    451120: {
        "branche": "Detailhandel med personbiler, varebiler og minibusser",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 45001,
        "grp127hovedafdeling": "Bilhandel",
    },
    451910: {
        "branche": "Engros- og detailhandel med campingkøretøjer, små trailere mv.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 45001,
        "grp127hovedafdeling": "Bilhandel",
    },
    451920: {
        "branche": "Engros- og detailhandel med lastbiler og påhængsvogne mv.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 45001,
        "grp127hovedafdeling": "Bilhandel",
    },
    452010: {
        "branche": "Autoreparationsværksteder mv.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 45002,
        "grp127hovedafdeling": "Bilværksteder mv.",
    },
    452020: {
        "branche": "Karosseriværksteder og autolakererier",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 45002,
        "grp127hovedafdeling": "Bilværksteder mv.",
    },
    452030: {
        "branche": "Undervognsbehandling",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 45002,
        "grp127hovedafdeling": "Bilværksteder mv.",
    },
    452040: {
        "branche": "Dækservice",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 45002,
        "grp127hovedafdeling": "Bilværksteder mv.",
    },
    453100: {
        "branche": "Engroshandel med reservedele og tilbehør til motorkøretøjer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 45002,
        "grp127hovedafdeling": "Bilværksteder mv.",
    },
    453200: {
        "branche": "Detailhandel med reservedele og tilbehør til motorkøretøjer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 45002,
        "grp127hovedafdeling": "Bilværksteder mv.",
    },
    454000: {
        "branche": "Salg, vedligeholdelse og reparation af motorcykler og reservedele og tilbehør hertil",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 45002,
        "grp127hovedafdeling": "Bilværksteder mv.",
    },
    461100: {
        "branche": "Agenturhandel med landbrugsråvarer, levende dyr, tekstilmaterialer og halvfabrikata",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46001,
        "grp127hovedafdeling": "Agenturhandel",
    },
    461200: {
        "branche": "Agenturhandel med brændstoffer, malme, metaller og kemiske produkter til industrien",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46001,
        "grp127hovedafdeling": "Agenturhandel",
    },
    461300: {
        "branche": "Agenturhandel med tømmer og andre byggematerialer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46001,
        "grp127hovedafdeling": "Agenturhandel",
    },
    461400: {
        "branche": "Agenturhandel med maskiner, teknisk udstyr, skibe og flyvemaskiner",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46001,
        "grp127hovedafdeling": "Agenturhandel",
    },
    461500: {
        "branche": "Agenturhandel med møbler, husholdningsartikler og isenkram",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46001,
        "grp127hovedafdeling": "Agenturhandel",
    },
    461600: {
        "branche": "Agenturhandel med tekstiler, beklædning, pelsværk, fodtøj og lædervarer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46001,
        "grp127hovedafdeling": "Agenturhandel",
    },
    461710: {
        "branche": "Fiskeauktioner",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46001,
        "grp127hovedafdeling": "Agenturhandel",
    },
    461790: {
        "branche": "Anden agenturhandel med føde-, drikke- og tobaksvarer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46001,
        "grp127hovedafdeling": "Agenturhandel",
    },
    461800: {
        "branche": "Agenturhandel med specialiseret varesortiment i.a.n.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46001,
        "grp127hovedafdeling": "Agenturhandel",
    },
    461900: {
        "branche": "Agenturhandel med blandet sortiment",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46001,
        "grp127hovedafdeling": "Agenturhandel",
    },
    462100: {
        "branche": "Engroshandel med korn, uforarbejdet tobak, såsæd og foderstoffer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46002,
        "grp127hovedafdeling": "Engrosh. med korn og foderstof",
    },
    462200: {
        "branche": "Engroshandel med blomster og planter",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46002,
        "grp127hovedafdeling": "Engrosh. med korn og foderstof",
    },
    462300: {
        "branche": "Engroshandel med levende dyr",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46002,
        "grp127hovedafdeling": "Engrosh. med korn og foderstof",
    },
    462400: {
        "branche": "Engroshandel med huder, skind og læder",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46002,
        "grp127hovedafdeling": "Engrosh. med korn og foderstof",
    },
    463100: {
        "branche": "Engroshandel med frugt og grøntsager",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46003,
        "grp127hovedafdeling": "Engrosh. med føde-, drikke- og tobaksvarer",
    },
    463200: {
        "branche": "Engroshandel med kød og kødprodukter",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46003,
        "grp127hovedafdeling": "Engrosh. med føde-, drikke- og tobaksvarer",
    },
    463300: {
        "branche": "Engroshandel med mejeriprodukter, æg samt spiselige olier og fedtstoffer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46003,
        "grp127hovedafdeling": "Engrosh. med føde-, drikke- og tobaksvarer",
    },
    463410: {
        "branche": "Engroshandel med øl, mineralvand, frugt- og grøntsagssaft",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46003,
        "grp127hovedafdeling": "Engrosh. med føde-, drikke- og tobaksvarer",
    },
    463420: {
        "branche": "Engroshandel med vin og spiritus",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46003,
        "grp127hovedafdeling": "Engrosh. med føde-, drikke- og tobaksvarer",
    },
    463500: {
        "branche": "Engroshandel med tobaksvarer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46003,
        "grp127hovedafdeling": "Engrosh. med føde-, drikke- og tobaksvarer",
    },
    463600: {
        "branche": "Engroshandel med sukker, chokolade og sukkervarer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46003,
        "grp127hovedafdeling": "Engrosh. med føde-, drikke- og tobaksvarer",
    },
    463700: {
        "branche": "Engroshandel med kaffe, te, kakao og krydderier",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46003,
        "grp127hovedafdeling": "Engrosh. med føde-, drikke- og tobaksvarer",
    },
    463810: {
        "branche": "Engroshandel med fisk og fiskeprodukter",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46003,
        "grp127hovedafdeling": "Engrosh. med føde-, drikke- og tobaksvarer",
    },
    463890: {
        "branche": "Specialiseret engroshandel med fødevarer i.a.n.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46003,
        "grp127hovedafdeling": "Engrosh. med føde-, drikke- og tobaksvarer",
    },
    463900: {
        "branche": "Ikke-specialiseret engroshandel med føde-, drikke- og tobaksvarer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46003,
        "grp127hovedafdeling": "Engrosh. med føde-, drikke- og tobaksvarer",
    },
    464100: {
        "branche": "Engroshandel med tekstiler",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464210: {
        "branche": "Engroshandel med beklædning",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464220: {
        "branche": "Engroshandel med fodtøj",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464310: {
        "branche": "Engroshandel med hårde hvidevarer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464320: {
        "branche": "Engroshandel med radio og tv mv.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464330: {
        "branche": "Engroshandel med indspillede videoer, cd'er, dvd'er mv.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464340: {
        "branche": "Engroshandel med fotografiske og optiske artikler",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464350: {
        "branche": "Engroshandel med elektriske husholdningsartikler",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464410: {
        "branche": "Engroshandel med porcelæns- og glasvarer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464420: {
        "branche": "Engroshandel med rengøringsmidler",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464500: {
        "branche": "Engroshandel med parfumerivarer og kosmetik",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464610: {
        "branche": "Engroshandel med medicinalvarer og sygeplejeartikler",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464620: {
        "branche": "Engroshandel med læge- og hospitalsartikler",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464700: {
        "branche": "Engroshandel med møbler, tæpper og belysningsartikler",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464800: {
        "branche": "Engroshandel med ure, smykker og guld- og sølvvarer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464910: {
        "branche": "Engroshandel med cykler, sportsartikler og lystbåde",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464920: {
        "branche": "Engroshandel med bøger, papir og papirvarer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464930: {
        "branche": "Engroshandel med kufferter og lædervarer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    464990: {
        "branche": "Engroshandel med andre husholdningsartikler",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46004,
        "grp127hovedafdeling": "Engrosh. med tekstiler og husholdningsudstyr",
    },
    465100: {
        "branche": "Engroshandel med computere, ydre enheder og software",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46005,
        "grp127hovedafdeling": "Engrosh. med it-udstyr",
    },
    465210: {
        "branche": "Engroshandel med elektronisk udstyr",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46005,
        "grp127hovedafdeling": "Engrosh. med it-udstyr",
    },
    465220: {
        "branche": "Engroshandel med telekommunikationsudstyr",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46005,
        "grp127hovedafdeling": "Engrosh. med it-udstyr",
    },
    466100: {
        "branche": "Engroshandel med landbrugsmaskiner, -udstyr og tilbehør hertil",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46006,
        "grp127hovedafdeling": "Engrosh. med andre maskiner",
    },
    466200: {
        "branche": "Engroshandel med værktøjsmaskiner",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46006,
        "grp127hovedafdeling": "Engrosh. med andre maskiner",
    },
    466300: {
        "branche": "Engroshandel med maskiner til minedrift og bygge- og anlægsvirksomhed",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46006,
        "grp127hovedafdeling": "Engrosh. med andre maskiner",
    },
    466400: {
        "branche": "Engroshandel med maskiner, udstyr og tilbehør til tekstilindustrien",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46006,
        "grp127hovedafdeling": "Engrosh. med andre maskiner",
    },
    466500: {
        "branche": "Engroshandel med kontormøbler",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46006,
        "grp127hovedafdeling": "Engrosh. med andre maskiner",
    },
    466600: {
        "branche": "Engroshandel med andre kontormaskiner og andet kontorudstyr",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46006,
        "grp127hovedafdeling": "Engrosh. med andre maskiner",
    },
    466900: {
        "branche": "Engroshandel med andre maskiner og andet udstyr",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46006,
        "grp127hovedafdeling": "Engrosh. med andre maskiner",
    },
    467100: {
        "branche": "Engroshandel med fast, flydende og luftformigt brændstof og lignende varer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46007,
        "grp127hovedafdeling": "Anden engroshandel",
    },
    467200: {
        "branche": "Engroshandel med metaller og metalmalme",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46007,
        "grp127hovedafdeling": "Anden engroshandel",
    },
    467310: {
        "branche": "Engroshandel med træ, trælast og byggematerialer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46007,
        "grp127hovedafdeling": "Anden engroshandel",
    },
    467320: {
        "branche": "Engroshandel med lak, maling, tapet, gulvbelægning mv.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46007,
        "grp127hovedafdeling": "Anden engroshandel",
    },
    467400: {
        "branche": "Engroshandel med isenkram, varmeanlæg og tilbehør",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46007,
        "grp127hovedafdeling": "Anden engroshandel",
    },
    467500: {
        "branche": "Engroshandel med kemiske produkter",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46007,
        "grp127hovedafdeling": "Anden engroshandel",
    },
    467600: {
        "branche": "Engroshandel med andre råvarer og halvfabrikata",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46007,
        "grp127hovedafdeling": "Anden engroshandel",
    },
    467700: {
        "branche": "Engroshandel med affaldsprodukter",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46007,
        "grp127hovedafdeling": "Anden engroshandel",
    },
    469000: {
        "branche": "Ikke-specialiseret engroshandel",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 46007,
        "grp127hovedafdeling": "Anden engroshandel",
    },
    471110: {
        "branche": "Købmænd og døgnkiosker",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47001,
        "grp127hovedafdeling": "Supermarkeder og varehuse mv.",
    },
    471120: {
        "branche": "Supermarkeder",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47001,
        "grp127hovedafdeling": "Supermarkeder og varehuse mv.",
    },
    471130: {
        "branche": "Discountforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47001,
        "grp127hovedafdeling": "Supermarkeder og varehuse mv.",
    },
    471900: {
        "branche": "Anden detailhandel fra ikke-specialiserede forretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47001,
        "grp127hovedafdeling": "Supermarkeder og varehuse mv.",
    },
    472100: {
        "branche": "Frugt- og grøntforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47002,
        "grp127hovedafdeling": "Specialbutikker med fødevarer - Specialbutikker med fødevarer ",
    },
    472200: {
        "branche": "Slagter- og viktualieforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47002,
        "grp127hovedafdeling": "Specialbutikker med fødevarer - Specialbutikker med fødevarer ",
    },
    472300: {
        "branche": "Fiskeforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47002,
        "grp127hovedafdeling": "Specialbutikker med fødevarer - Specialbutikker med fødevarer ",
    },
    472400: {
        "branche": "Detailhandel med brød, konditori- og sukkervarer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47002,
        "grp127hovedafdeling": "Specialbutikker med fødevarer - Specialbutikker med fødevarer ",
    },
    472500: {
        "branche": "Detailhandel med drikkevarer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47002,
        "grp127hovedafdeling": "Specialbutikker med fødevarer - Specialbutikker med fødevarer ",
    },
    472600: {
        "branche": "Tobaksforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47002,
        "grp127hovedafdeling": "Specialbutikker med fødevarer - Specialbutikker med fødevarer ",
    },
    472900: {
        "branche": "Anden detailhandel med fødevarer i specialforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47002,
        "grp127hovedafdeling": "Specialbutikker med fødevarer - Specialbutikker med fødevarer ",
    },
    473000: {
        "branche": "Servicestationer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47003,
        "grp127hovedafdeling": "Tankstationer",
    },
    474100: {
        "branche": "Detailhandel med computere, ydre enheder og software",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47004,
        "grp127hovedafdeling": "Detailh. med forbrugerelektronik",
    },
    474200: {
        "branche": "Detailhandel med telekommunikationsudstyr",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47004,
        "grp127hovedafdeling": "Detailh. med forbrugerelektronik",
    },
    474300: {
        "branche": "Radio- og tv-forretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47004,
        "grp127hovedafdeling": "Detailh. med forbrugerelektronik",
    },
    475100: {
        "branche": "Detailhandel med kjolestoffer, garn, broderier mv.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    475210: {
        "branche": "Farve- og tapetforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    475220: {
        "branche": "Byggemarkeder og værktøjsmagasiner",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    475300: {
        "branche": "Detailhandel med tæpper, vægbeklædning og gulvbelægning",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    475400: {
        "branche": "Detailhandel med elektriske husholdningsapparater",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    475910: {
        "branche": "Møbelforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    475920: {
        "branche": "Boligtekstilforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    475930: {
        "branche": "Detailhandel med køkkenudstyr, glas, porcelæn, bestik, vaser, lysestager mv.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    475940: {
        "branche": "Forhandlere af musikinstrumenter",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    475990: {
        "branche": "Detailhandel med møbler og belysningsartikler samt husholdningsartikler i.a.n.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    476100: {
        "branche": "Detailhandel med bøger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47006,
        "grp127hovedafdeling": "Detailh. med kultur- og fritidsprodukter",
    },
    476200: {
        "branche": "Detailhandel med aviser og papirvarer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47006,
        "grp127hovedafdeling": "Detailh. med kultur- og fritidsprodukter",
    },
    476300: {
        "branche": "Detailhandel med musik- og videooptagelser",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47006,
        "grp127hovedafdeling": "Detailh. med kultur- og fritidsprodukter",
    },
    476410: {
        "branche": "Forhandlere af sports- og campingudstyr",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47006,
        "grp127hovedafdeling": "Detailh. med kultur- og fritidsprodukter",
    },
    476420: {
        "branche": "Cykel- og knallertforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47006,
        "grp127hovedafdeling": "Detailh. med kultur- og fritidsprodukter",
    },
    476430: {
        "branche": "Forhandlere af lystbåde og udstyr hertil",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47006,
        "grp127hovedafdeling": "Detailh. med kultur- og fritidsprodukter",
    },
    476500: {
        "branche": "Detailhandel med spil og legetøj",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47006,
        "grp127hovedafdeling": "Detailh. med kultur- og fritidsprodukter",
    },
    477110: {
        "branche": "Tøjforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47007,
        "grp127hovedafdeling": "Detailh. med beklædning og fodtøj",
    },
    477120: {
        "branche": "Babyudstyrs- og børnetøjsforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47007,
        "grp127hovedafdeling": "Detailh. med beklædning og fodtøj",
    },
    477210: {
        "branche": "Skotøjsforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47007,
        "grp127hovedafdeling": "Detailh. med beklædning og fodtøj",
    },
    477220: {
        "branche": "Lædervareforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47007,
        "grp127hovedafdeling": "Detailh. med beklædning og fodtøj",
    },
    477300: {
        "branche": "Apoteker",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    477400: {
        "branche": "Detailhandel med medicinske og ortopædiske artikler",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    477500: {
        "branche": "Detailhandel med kosmetikvarer og produkter til personlig pleje",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    477610: {
        "branche": "Blomsterforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    477620: {
        "branche": "Planteforhandlere og havecentre",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    477630: {
        "branche": "Dyrehandel",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    477700: {
        "branche": "Detailhandel med ure, smykker og guld- og sølvvarer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    477810: {
        "branche": "Optikere",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    477820: {
        "branche": "Fotoforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    477830: {
        "branche": "Forhandlere af gaveartikler og brugskunst",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    477840: {
        "branche": "Kunsthandel og gallerivirksomhed",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    477890: {
        "branche": "Detailhandel med andre varer i.a.n.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    477900: {
        "branche": "Detailhandel med brugte varer i forretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47005,
        "grp127hovedafdeling": "Detailh. med tekstiler og husholdningsudstyr mv.",
    },
    478100: {
        "branche": "Detailhandel med føde-, drikke- og tobaksvarer fra stadepladser og markeder",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47008,
        "grp127hovedafdeling": "Internethandel, postordre mv.",
    },
    478200: {
        "branche": "Detailhandel med tekstiler, beklædningsartikler og fodtøj fra stadepladser og markeder",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47008,
        "grp127hovedafdeling": "Internethandel, postordre mv.",
    },
    478900: {
        "branche": "Detailhandel med andre varer fra stadepladser og markeder",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47008,
        "grp127hovedafdeling": "Internethandel, postordre mv.",
    },
    479111: {
        "branche": "Detailhandel med dagligvarer via internet",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47008,
        "grp127hovedafdeling": "Internethandel, postordre mv.",
    },
    479112: {
        "branche": "Detailhandel med elektroniske eller elektriske apparater samt fotoudstyr via internet",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47008,
        "grp127hovedafdeling": "Internethandel, postordre mv.",
    },
    479113: {
        "branche": "Detailhandel med husholdnings- eller boligudstyr, bortset fra elektriske apparater, via internet",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47008,
        "grp127hovedafdeling": "Internethandel, postordre mv.",
    },
    479114: {
        "branche": "Detailhandel med bøger, kontorartikler, musik eller film via internet",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47008,
        "grp127hovedafdeling": "Internethandel, postordre mv.",
    },
    479115: {
        "branche": "Detailhandel med hobbyartikler, musikinstrumenter, sportsudstyr, legetøj, cykler via internet",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47008,
        "grp127hovedafdeling": "Internethandel, postordre mv.",
    },
    479116: {
        "branche": "Detailhandel med tøj, sko, lædervarer, ure eller babyudstyr  via internet",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47008,
        "grp127hovedafdeling": "Internethandel, postordre mv.",
    },
    479117: {
        "branche": "Detailhandel med medicin og produkter til personlig pleje via internet",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47008,
        "grp127hovedafdeling": "Internethandel, postordre mv.",
    },
    479119: {
        "branche": "Detailhandel med andre varer i.a.n. via internet",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47008,
        "grp127hovedafdeling": "Internethandel, postordre mv.",
    },
    479120: {
        "branche": "Detailhandel via internettet med digitale produkter",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47008,
        "grp127hovedafdeling": "Internethandel, postordre mv.",
    },
    479900: {
        "branche": "Anden detailhandel undtagen fra forretninger, stadepladser og markeder",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "G",
        "grp19hovedafdeling": "Handel - Handel ",
        "grp36kode": "G",
        "grp36hovedafdeling": "Handel",
        "grp127kode": 47008,
        "grp127hovedafdeling": "Internethandel, postordre mv.",
    },
    491000: {
        "branche": "Passagertransport med regional- eller fjerntog",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 49001,
        "grp127hovedafdeling": "Regional- og fjerntog",
    },
    492000: {
        "branche": "Godstransport med tog",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 49001,
        "grp127hovedafdeling": "Regional- og fjerntog",
    },
    493110: {
        "branche": "Rutebuskørsel, by- og nærtrafik",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 49002,
        "grp127hovedafdeling": "Lokaltog, bus og taxi mv.",
    },
    493120: {
        "branche": "S-togstrafik, metro og andre nærbaner",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 49002,
        "grp127hovedafdeling": "Lokaltog, bus og taxi mv.",
    },
    493200: {
        "branche": "Taxikørsel",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 49002,
        "grp127hovedafdeling": "Lokaltog, bus og taxi mv.",
    },
    493910: {
        "branche": "Rutebuskørsel, fjerntrafik og skolebusser",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 49002,
        "grp127hovedafdeling": "Lokaltog, bus og taxi mv.",
    },
    493920: {
        "branche": "Turistkørsel og anden landpassagertransport",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 49002,
        "grp127hovedafdeling": "Lokaltog, bus og taxi mv.",
    },
    494100: {
        "branche": "Vejgodstransport",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 49003,
        "grp127hovedafdeling": "Fragtvognmænd og rørtransport",
    },
    494200: {
        "branche": "Flytteforretninger",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 49003,
        "grp127hovedafdeling": "Fragtvognmænd og rørtransport",
    },
    495000: {
        "branche": "Rørtransport",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 49003,
        "grp127hovedafdeling": "Fragtvognmænd og rørtransport",
    },
    501000: {
        "branche": "Sø- og kysttransport af passagerer",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 50000,
        "grp127hovedafdeling": "Skibsfart",
    },
    502000: {
        "branche": "Sø- og kysttransport af gods",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 50000,
        "grp127hovedafdeling": "Skibsfart",
    },
    503000: {
        "branche": "Transport af passagerer ad indre vandveje",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 50000,
        "grp127hovedafdeling": "Skibsfart",
    },
    504000: {
        "branche": "Transport af gods ad indre vandveje",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 50000,
        "grp127hovedafdeling": "Skibsfart",
    },
    511010: {
        "branche": "Ruteflyvning",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 51000,
        "grp127hovedafdeling": "Luftfart",
    },
    511020: {
        "branche": "Charter- og taxiflyvning",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 51000,
        "grp127hovedafdeling": "Luftfart",
    },
    512100: {
        "branche": "Lufttransport af gods",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 51000,
        "grp127hovedafdeling": "Luftfart",
    },
    512200: {
        "branche": "Rumfart",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 51000,
        "grp127hovedafdeling": "Luftfart",
    },
    521000: {
        "branche": "Oplagrings- og pakhusvirksomhed",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 52000,
        "grp127hovedafdeling": "Hjælpevirksomhed til transport",
    },
    522110: {
        "branche": "Stationer, godsterminaler mv.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 52000,
        "grp127hovedafdeling": "Hjælpevirksomhed til transport",
    },
    522120: {
        "branche": "Parkering og vejhjælp mv.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 52000,
        "grp127hovedafdeling": "Hjælpevirksomhed til transport",
    },
    522130: {
        "branche": "Drift af betalingsveje, -broer og -tunneler",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 52000,
        "grp127hovedafdeling": "Hjælpevirksomhed til transport",
    },
    522210: {
        "branche": "Erhvervshavne",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 52000,
        "grp127hovedafdeling": "Hjælpevirksomhed til transport",
    },
    522220: {
        "branche": "Bugserings-, bjærgnings- og redningsvæsen mv.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 52000,
        "grp127hovedafdeling": "Hjælpevirksomhed til transport",
    },
    522300: {
        "branche": "Serviceydelser i forbindelse med luftfart",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 52000,
        "grp127hovedafdeling": "Hjælpevirksomhed til transport",
    },
    522400: {
        "branche": "Godshåndtering",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 52000,
        "grp127hovedafdeling": "Hjælpevirksomhed til transport",
    },
    522910: {
        "branche": "Skibsmæglere",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 52000,
        "grp127hovedafdeling": "Hjælpevirksomhed til transport",
    },
    522920: {
        "branche": "Speditører",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 52000,
        "grp127hovedafdeling": "Hjælpevirksomhed til transport",
    },
    522990: {
        "branche": "Andre tjenesteydelser i forbindelse med transport",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 52000,
        "grp127hovedafdeling": "Hjælpevirksomhed til transport",
    },
    531000: {
        "branche": "Posttjenester omfattet af forsyningspligten",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 53000,
        "grp127hovedafdeling": "Post og kurertjeneste",
    },
    532000: {
        "branche": "Andre post- og kurertjenester",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "H",
        "grp19hovedafdeling": "Transport",
        "grp36kode": "H",
        "grp36hovedafdeling": "Transport",
        "grp127kode": 53000,
        "grp127hovedafdeling": "Post og kurertjeneste",
    },
    551010: {
        "branche": "Hoteller",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "I",
        "grp19hovedafdeling": "Hoteller og restauranter",
        "grp36kode": "I",
        "grp36hovedafdeling": "Hoteller og restauranter - Hoteller og restauranter   ",
        "grp127kode": 55000,
        "grp127hovedafdeling": "Hoteller mv.",
    },
    551020: {
        "branche": "Konferencecentre og kursusejendomme",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "I",
        "grp19hovedafdeling": "Hoteller og restauranter",
        "grp36kode": "I",
        "grp36hovedafdeling": "Hoteller og restauranter - Hoteller og restauranter   ",
        "grp127kode": 55000,
        "grp127hovedafdeling": "Hoteller mv.",
    },
    552000: {
        "branche": "Ferieboliger og andre indlogeringsfaciliteter til kortvarige ophold",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "I",
        "grp19hovedafdeling": "Hoteller og restauranter",
        "grp36kode": "I",
        "grp36hovedafdeling": "Hoteller og restauranter - Hoteller og restauranter   ",
        "grp127kode": 55000,
        "grp127hovedafdeling": "Hoteller mv.",
    },
    553000: {
        "branche": "Campingpladser",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "I",
        "grp19hovedafdeling": "Hoteller og restauranter",
        "grp36kode": "I",
        "grp36hovedafdeling": "Hoteller og restauranter - Hoteller og restauranter   ",
        "grp127kode": 55000,
        "grp127hovedafdeling": "Hoteller mv.",
    },
    559000: {
        "branche": "Andre overnatningsfaciliteter",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "I",
        "grp19hovedafdeling": "Hoteller og restauranter",
        "grp36kode": "I",
        "grp36hovedafdeling": "Hoteller og restauranter - Hoteller og restauranter   ",
        "grp127kode": 55000,
        "grp127hovedafdeling": "Hoteller mv.",
    },
    561010: {
        "branche": "Restauranter",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "I",
        "grp19hovedafdeling": "Hoteller og restauranter",
        "grp36kode": "I",
        "grp36hovedafdeling": "Hoteller og restauranter - Hoteller og restauranter   ",
        "grp127kode": 56000,
        "grp127hovedafdeling": "Restauranter",
    },
    561020: {
        "branche": "Pizzeriaer, grillbarer, isbarer mv.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "I",
        "grp19hovedafdeling": "Hoteller og restauranter",
        "grp36kode": "I",
        "grp36hovedafdeling": "Hoteller og restauranter - Hoteller og restauranter   ",
        "grp127kode": 56000,
        "grp127hovedafdeling": "Restauranter",
    },
    562100: {
        "branche": "Event catering",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "I",
        "grp19hovedafdeling": "Hoteller og restauranter",
        "grp36kode": "I",
        "grp36hovedafdeling": "Hoteller og restauranter - Hoteller og restauranter   ",
        "grp127kode": 56000,
        "grp127hovedafdeling": "Restauranter",
    },
    562900: {
        "branche": "Anden restaurationsvirksomhed",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "I",
        "grp19hovedafdeling": "Hoteller og restauranter",
        "grp36kode": "I",
        "grp36hovedafdeling": "Hoteller og restauranter - Hoteller og restauranter   ",
        "grp127kode": 56000,
        "grp127hovedafdeling": "Restauranter",
    },
    563000: {
        "branche": "Caféer, værtshuse, diskoteker mv.",
        "grp10kode": 4,
        "grp10hovedafdeling": "Handel og transport",
        "grp19kode": "I",
        "grp19hovedafdeling": "Hoteller og restauranter",
        "grp36kode": "I",
        "grp36hovedafdeling": "Hoteller og restauranter - Hoteller og restauranter   ",
        "grp127kode": 56000,
        "grp127hovedafdeling": "Restauranter",
    },
    581100: {
        "branche": "Udgivelse af bøger",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 58001,
        "grp127hovedafdeling": "Forlag",
    },
    581200: {
        "branche": "Udgivelse af telefonbøger og adresselister",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 58001,
        "grp127hovedafdeling": "Forlag",
    },
    581300: {
        "branche": "Udgivelse af aviser og dagblade",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 58001,
        "grp127hovedafdeling": "Forlag",
    },
    581410: {
        "branche": "Udgivelse af ugeblade og magasiner",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 58001,
        "grp127hovedafdeling": "Forlag",
    },
    581420: {
        "branche": "Udgivelse af distrikts- og annonceblade",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 58001,
        "grp127hovedafdeling": "Forlag",
    },
    581900: {
        "branche": "Anden udgivervirksomhed",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 58001,
        "grp127hovedafdeling": "Forlag",
    },
    582100: {
        "branche": "Udgivelse af computerspil",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 58002,
        "grp127hovedafdeling": "Udgivelse af computerspil og anden software",
    },
    582900: {
        "branche": "Anden udgivelse af software",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 58002,
        "grp127hovedafdeling": "Udgivelse af computerspil og anden software",
    },
    591110: {
        "branche": "Produktion af film og videofilm",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 59000,
        "grp127hovedafdeling": "Udgivelse af computerspil og anden software",
    },
    591120: {
        "branche": "Produktion af tv-programmer",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 59000,
        "grp127hovedafdeling": "Udgivelse af computerspil og anden software",
    },
    591200: {
        "branche": "Aktiviteter, der udøves efter produktion af film, video- og tv-programmer",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 59000,
        "grp127hovedafdeling": "Udgivelse af computerspil og anden software",
    },
    591300: {
        "branche": "Distribution af film, video- og tv-programmer",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 59000,
        "grp127hovedafdeling": "Udgivelse af computerspil og anden software",
    },
    591400: {
        "branche": "Biografer",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 59000,
        "grp127hovedafdeling": "Udgivelse af computerspil og anden software",
    },
    592000: {
        "branche": "Indspilning af lydoptagelser og udgivelse af musik",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 59000,
        "grp127hovedafdeling": "Udgivelse af computerspil og anden software",
    },
    601000: {
        "branche": "Radiovirksomhed",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 60000,
        "grp127hovedafdeling": "Radio- og tv-stationer",
    },
    602000: {
        "branche": "Tv-virksomhed",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JA",
        "grp36hovedafdeling": "Forlag, tv og radio - Forlag, tv og radio   ",
        "grp127kode": 60000,
        "grp127hovedafdeling": "Radio- og tv-stationer",
    },
    611000: {
        "branche": "Fastnetbaseret telekommunikation",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JB",
        "grp36hovedafdeling": "Telekommunikation - Telekommunikation   ",
        "grp127kode": 61000,
        "grp127hovedafdeling": "Telekommunikation",
    },
    612000: {
        "branche": "Trådløs telekommunikation",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JB",
        "grp36hovedafdeling": "Telekommunikation - Telekommunikation   ",
        "grp127kode": 61000,
        "grp127hovedafdeling": "Telekommunikation",
    },
    613000: {
        "branche": "Satellitbaseret telekommunikation",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JB",
        "grp36hovedafdeling": "Telekommunikation - Telekommunikation   ",
        "grp127kode": 61000,
        "grp127hovedafdeling": "Telekommunikation",
    },
    619000: {
        "branche": "Anden telekommunikation",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JB",
        "grp36hovedafdeling": "Telekommunikation - Telekommunikation   ",
        "grp127kode": 61000,
        "grp127hovedafdeling": "Telekommunikation",
    },
    620100: {
        "branche": "Computerprogrammering",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JC",
        "grp36hovedafdeling": "It- og informationstjenester - It- og informationstjenester   ",
        "grp127kode": 62000,
        "grp127hovedafdeling": "It-konsulenter mv.",
    },
    620200: {
        "branche": "Konsulentbistand vedrørende informationsteknologi",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JC",
        "grp36hovedafdeling": "It- og informationstjenester - It- og informationstjenester   ",
        "grp127kode": 62000,
        "grp127hovedafdeling": "It-konsulenter mv.",
    },
    620300: {
        "branche": "Computer facility management",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JC",
        "grp36hovedafdeling": "It- og informationstjenester - It- og informationstjenester   ",
        "grp127kode": 62000,
        "grp127hovedafdeling": "It-konsulenter mv.",
    },
    620900: {
        "branche": "Anden it-servicevirksomhed",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JC",
        "grp36hovedafdeling": "It- og informationstjenester - It- og informationstjenester   ",
        "grp127kode": 62000,
        "grp127hovedafdeling": "It-konsulenter mv.",
    },
    631100: {
        "branche": "Databehandling, webhosting og lignende serviceydelser",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JC",
        "grp36hovedafdeling": "It- og informationstjenester - It- og informationstjenester   ",
        "grp127kode": 63000,
        "grp127hovedafdeling": "Informationstjenester",
    },
    631200: {
        "branche": "Webportaler",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JC",
        "grp36hovedafdeling": "It- og informationstjenester - It- og informationstjenester   ",
        "grp127kode": 63000,
        "grp127hovedafdeling": "Informationstjenester",
    },
    639100: {
        "branche": "Pressebureauer",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JC",
        "grp36hovedafdeling": "It- og informationstjenester - It- og informationstjenester   ",
        "grp127kode": 63000,
        "grp127hovedafdeling": "Informationstjenester",
    },
    639900: {
        "branche": "Andre informationstjenester i.a.n.",
        "grp10kode": 5,
        "grp10hovedafdeling": "Information og kommunikation",
        "grp19kode": "J",
        "grp19hovedafdeling": "Information og kommunikation",
        "grp36kode": "JC",
        "grp36hovedafdeling": "It- og informationstjenester - It- og informationstjenester   ",
        "grp127kode": 63000,
        "grp127hovedafdeling": "Informationstjenester",
    },
    641100: {
        "branche": "Centralbanker",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64001,
        "grp127hovedafdeling": "Pengeinstitutter",
    },
    641900: {
        "branche": "Banker, sparekasser og andelskasser",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64001,
        "grp127hovedafdeling": "Pengeinstitutter",
    },
    642010: {
        "branche": "Finansielle holdingselskaber",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64002,
        "grp127hovedafdeling": "Kreditforeninger mv.",
    },
    642020: {
        "branche": "Ikke-finansielle holdingselskaber",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64002,
        "grp127hovedafdeling": "Kreditforeninger mv.",
    },
    642030: {
        "branche": "Gennemløbsholdingselskaber",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64002,
        "grp127hovedafdeling": "Kreditforeninger mv.",
    },
    643010: {
        "branche": "Investeringsforeninger",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64002,
        "grp127hovedafdeling": "Kreditforeninger mv.",
    },
    643020: {
        "branche": "Pengemarkedsforeninger",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64002,
        "grp127hovedafdeling": "Kreditforeninger mv.",
    },
    643030: {
        "branche": "Investeringsselskaber",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64002,
        "grp127hovedafdeling": "Kreditforeninger mv.",
    },
    643040: {
        "branche": "Ventureselskaber og kapitalfonde",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64002,
        "grp127hovedafdeling": "Kreditforeninger mv.",
    },
    649100: {
        "branche": "Finansiel leasing",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64002,
        "grp127hovedafdeling": "Kreditforeninger mv.",
    },
    649210: {
        "branche": "Realkreditinstitutter",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64002,
        "grp127hovedafdeling": "Kreditforeninger mv.",
    },
    649220: {
        "branche": "Andre kreditinstitutter",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64002,
        "grp127hovedafdeling": "Kreditforeninger mv.",
    },
    649230: {
        "branche": "Andre kreditselskaber",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64002,
        "grp127hovedafdeling": "Kreditforeninger mv.",
    },
    649240: {
        "branche": "FVC-selskaber",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64002,
        "grp127hovedafdeling": "Kreditforeninger mv.",
    },
    649900: {
        "branche": "Anden finansiel formidling undtagen forsikring og pensionsforsikring, i.a.n.",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 64002,
        "grp127hovedafdeling": "Kreditforeninger mv.",
    },
    651100: {
        "branche": "Livsforsikring",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 65000,
        "grp127hovedafdeling": "Forsikring og pension",
    },
    651200: {
        "branche": "Anden forsikring",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 65000,
        "grp127hovedafdeling": "Forsikring og pension",
    },
    652000: {
        "branche": "Genforsikring",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 65000,
        "grp127hovedafdeling": "Forsikring og pension",
    },
    653010: {
        "branche": "Pensionskasser",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 65000,
        "grp127hovedafdeling": "Forsikring og pension",
    },
    653020: {
        "branche": "Anden pensionsforsikring",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 65000,
        "grp127hovedafdeling": "Forsikring og pension",
    },
    661100: {
        "branche": "Forvaltning af kapitalmarkeder",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 66000,
        "grp127hovedafdeling": "Finansiel service",
    },
    661200: {
        "branche": "Værdipapir- og varemægling",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 66000,
        "grp127hovedafdeling": "Finansiel service",
    },
    661900: {
        "branche": "Andre hjælpetjenester i forbindelse med finansiel formidling",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 66000,
        "grp127hovedafdeling": "Finansiel service",
    },
    662100: {
        "branche": "Risiko- og skadesvurdering",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 66000,
        "grp127hovedafdeling": "Finansiel service",
    },
    662200: {
        "branche": "Forsikringsagenters og forsikringsmægleres virksomhed",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 66000,
        "grp127hovedafdeling": "Finansiel service",
    },
    662900: {
        "branche": "Andre hjælpetjenester i forbindelse med forsikring og pensionsforsikring",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 66000,
        "grp127hovedafdeling": "Finansiel service",
    },
    663000: {
        "branche": "Formueforvaltning",
        "grp10kode": 6,
        "grp10hovedafdeling": "Finansiering og forsikring",
        "grp19kode": "K",
        "grp19hovedafdeling": "Finansiering og forsikring",
        "grp36kode": "K",
        "grp36hovedafdeling": "Finansiering og forsikring - Finansiering og forsikring   ",
        "grp127kode": 66000,
        "grp127hovedafdeling": "Finansiel service",
    },
    681000: {
        "branche": "Køb og salg af egen fast ejendom",
        "grp10kode": 7,
        "grp10hovedafdeling": "Ejendomshandel og udlejning",
        "grp19kode": "L",
        "grp19hovedafdeling": "Ejendomshandel og udlejning",
        "grp36kode": "L",
        "grp36hovedafdeling": "Ejendomshandel og udlejning - Ejendomshandel og udlejning   ",
        "grp127kode": 68001,
        "grp127hovedafdeling": "Ejendomsmæglere mv.",
    },
    682010: {
        "branche": "Almennyttige boligselskaber",
        "grp10kode": 7,
        "grp10hovedafdeling": "Ejendomshandel og udlejning",
        "grp19kode": "L",
        "grp19hovedafdeling": "Ejendomshandel og udlejning",
        "grp36kode": "L",
        "grp36hovedafdeling": "Ejendomshandel og udlejning - Ejendomshandel og udlejning   ",
        "grp127kode": 68002,
        "grp127hovedafdeling": "Boligudlejning",
    },
    682020: {
        "branche": "Private andelsboligforeninger",
        "grp10kode": 7,
        "grp10hovedafdeling": "Ejendomshandel og udlejning",
        "grp19kode": "L",
        "grp19hovedafdeling": "Ejendomshandel og udlejning",
        "grp36kode": "L",
        "grp36hovedafdeling": "Ejendomshandel og udlejning - Ejendomshandel og udlejning   ",
        "grp127kode": 68002,
        "grp127hovedafdeling": "Boligudlejning",
    },
    682030: {
        "branche": "Anden udlejning af boliger",
        "grp10kode": 7,
        "grp10hovedafdeling": "Ejendomshandel og udlejning",
        "grp19kode": "L",
        "grp19hovedafdeling": "Ejendomshandel og udlejning",
        "grp36kode": "L",
        "grp36hovedafdeling": "Ejendomshandel og udlejning - Ejendomshandel og udlejning   ",
        "grp127kode": 68002,
        "grp127hovedafdeling": "Boligudlejning",
    },
    682040: {
        "branche": "Udlejning af erhvervsejendomme",
        "grp10kode": 7,
        "grp10hovedafdeling": "Ejendomshandel og udlejning",
        "grp19kode": "L",
        "grp19hovedafdeling": "Ejendomshandel og udlejning",
        "grp36kode": "L",
        "grp36hovedafdeling": "Ejendomshandel og udlejning - Ejendomshandel og udlejning   ",
        "grp127kode": 68003,
        "grp127hovedafdeling": "Udlejning af erhvervsejendomme",
    },
    683110: {
        "branche": "Ejendomsmæglere",
        "grp10kode": 7,
        "grp10hovedafdeling": "Ejendomshandel og udlejning",
        "grp19kode": "L",
        "grp19hovedafdeling": "Ejendomshandel og udlejning",
        "grp36kode": "L",
        "grp36hovedafdeling": "Ejendomshandel og udlejning - Ejendomshandel og udlejning   ",
        "grp127kode": 68001,
        "grp127hovedafdeling": "Ejendomsmæglere mv.",
    },
    683120: {
        "branche": "Boliganvisning, ferieboligudlejning mv.",
        "grp10kode": 7,
        "grp10hovedafdeling": "Ejendomshandel og udlejning",
        "grp19kode": "L",
        "grp19hovedafdeling": "Ejendomshandel og udlejning",
        "grp36kode": "L",
        "grp36hovedafdeling": "Ejendomshandel og udlejning - Ejendomshandel og udlejning   ",
        "grp127kode": 68001,
        "grp127hovedafdeling": "Ejendomsmæglere mv.",
    },
    683210: {
        "branche": "Administration af fast ejendom på kontraktbasis",
        "grp10kode": 7,
        "grp10hovedafdeling": "Ejendomshandel og udlejning",
        "grp19kode": "L",
        "grp19hovedafdeling": "Ejendomshandel og udlejning",
        "grp36kode": "L",
        "grp36hovedafdeling": "Ejendomshandel og udlejning - Ejendomshandel og udlejning   ",
        "grp127kode": 68001,
        "grp127hovedafdeling": "Ejendomsmæglere mv.",
    },
    683220: {
        "branche": "Ejerforeninger",
        "grp10kode": 7,
        "grp10hovedafdeling": "Ejendomshandel og udlejning",
        "grp19kode": "L",
        "grp19hovedafdeling": "Ejendomshandel og udlejning",
        "grp36kode": "L",
        "grp36hovedafdeling": "Ejendomshandel og udlejning - Ejendomshandel og udlejning   ",
        "grp127kode": 68001,
        "grp127hovedafdeling": "Ejendomsmæglere mv.",
    },
    691000: {
        "branche": "Juridisk bistand",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 69001,
        "grp127hovedafdeling": "Advokatvirksomhed",
    },
    692000: {
        "branche": "Bogføring og revision; skatterådgivning",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 69002,
        "grp127hovedafdeling": "Revision og bogføring",
    },
    701010: {
        "branche": "Ikke-finansielle hovedsæders virksomhed",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 70000,
        "grp127hovedafdeling": "Virksomhedskonsulenter",
    },
    701020: {
        "branche": "Finansielle hovedsæders virksomhed",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 70000,
        "grp127hovedafdeling": "Virksomhedskonsulenter",
    },
    702100: {
        "branche": "Public relations og kommunikation",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 70000,
        "grp127hovedafdeling": "Virksomhedskonsulenter",
    },
    702200: {
        "branche": "Virksomhedsrådgivning og anden rådgivning om driftsledelse",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 70000,
        "grp127hovedafdeling": "Virksomhedskonsulenter",
    },
    711100: {
        "branche": "Arkitektvirksomhed",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 71000,
        "grp127hovedafdeling": "Arkitekter og rådgivende ingeniører",
    },
    711210: {
        "branche": "Rådgivende ingeniørvirksomhed inden for byggeri og anlægsarbejder",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 71000,
        "grp127hovedafdeling": "Arkitekter og rådgivende ingeniører",
    },
    711220: {
        "branche": "Rådgivende ingeniørvirksomhed inden for produktions- og maskinteknik",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 71000,
        "grp127hovedafdeling": "Arkitekter og rådgivende ingeniører",
    },
    711230: {
        "branche": "Opstilling og levering af færdige fabriksanlæg",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 71000,
        "grp127hovedafdeling": "Arkitekter og rådgivende ingeniører",
    },
    711240: {
        "branche": "Geologiske undersøgelser og prospektering, landinspektører mv.",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 71000,
        "grp127hovedafdeling": "Arkitekter og rådgivende ingeniører",
    },
    711290: {
        "branche": "Anden teknisk rådgivning",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 71000,
        "grp127hovedafdeling": "Arkitekter og rådgivende ingeniører",
    },
    712010: {
        "branche": "Kontrol af levnedsmidler",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 71000,
        "grp127hovedafdeling": "Arkitekter og rådgivende ingeniører",
    },
    712020: {
        "branche": "Teknisk afprøvning og kontrol",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 71000,
        "grp127hovedafdeling": "Arkitekter og rådgivende ingeniører",
    },
    712090: {
        "branche": "Anden måling og teknisk analyse",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MA",
        "grp36hovedafdeling": "Rådgivning mv. - Rådgivning mv.   ",
        "grp127kode": 71000,
        "grp127hovedafdeling": "Arkitekter og rådgivende ingeniører",
    },
    721100: {
        "branche": "Forskning og eksperimentel udvikling indenfor bioteknologi",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MB",
        "grp36hovedafdeling": "Forskning og udvikling - Forskning og udvikling   ",
        "grp127kode": 72000,
        "grp127hovedafdeling": "Forskning og udvikling",
    },
    721900: {
        "branche": "Anden forskning og eksperimentel udvikling inden for naturvidenskab og teknik",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MB",
        "grp36hovedafdeling": "Forskning og udvikling - Forskning og udvikling   ",
        "grp127kode": 72000,
        "grp127hovedafdeling": "Forskning og udvikling",
    },
    722000: {
        "branche": "Forskning og eksperimentel udvikling inden for samfundsvidenskab og humanistiske videnskaber",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MB",
        "grp36hovedafdeling": "Forskning og udvikling - Forskning og udvikling   ",
        "grp127kode": 72000,
        "grp127hovedafdeling": "Forskning og udvikling",
    },
    731110: {
        "branche": "Reklamebureauer",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MC",
        "grp36hovedafdeling": "Reklame og øvrig erhvervsservice - Reklame og øvrig erhvervsservice   ",
        "grp127kode": 73000,
        "grp127hovedafdeling": "Reklame- og analysebureauer",
    },
    731190: {
        "branche": "Anden reklamevirksomhed",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MC",
        "grp36hovedafdeling": "Reklame og øvrig erhvervsservice - Reklame og øvrig erhvervsservice   ",
        "grp127kode": 73000,
        "grp127hovedafdeling": "Reklame- og analysebureauer",
    },
    731200: {
        "branche": "Reklameplads i medier",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MC",
        "grp36hovedafdeling": "Reklame og øvrig erhvervsservice - Reklame og øvrig erhvervsservice   ",
        "grp127kode": 73000,
        "grp127hovedafdeling": "Reklame- og analysebureauer",
    },
    732000: {
        "branche": "Markedsanalyse og offentlig meningsmåling",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MC",
        "grp36hovedafdeling": "Reklame og øvrig erhvervsservice - Reklame og øvrig erhvervsservice   ",
        "grp127kode": 73000,
        "grp127hovedafdeling": "Reklame- og analysebureauer",
    },
    741010: {
        "branche": "Industriel design og produktdesign",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MC",
        "grp36hovedafdeling": "Reklame og øvrig erhvervsservice - Reklame og øvrig erhvervsservice   ",
        "grp127kode": 74000,
        "grp127hovedafdeling": "Anden videnservice",
    },
    741020: {
        "branche": "Kommunikationsdesign og grafisk design",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MC",
        "grp36hovedafdeling": "Reklame og øvrig erhvervsservice - Reklame og øvrig erhvervsservice   ",
        "grp127kode": 74000,
        "grp127hovedafdeling": "Anden videnservice",
    },
    741030: {
        "branche": "Indretningsarkitekter og rumdesign",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MC",
        "grp36hovedafdeling": "Reklame og øvrig erhvervsservice - Reklame og øvrig erhvervsservice   ",
        "grp127kode": 74000,
        "grp127hovedafdeling": "Anden videnservice",
    },
    742000: {
        "branche": "Fotografisk virksomhed",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MC",
        "grp36hovedafdeling": "Reklame og øvrig erhvervsservice - Reklame og øvrig erhvervsservice   ",
        "grp127kode": 74000,
        "grp127hovedafdeling": "Anden videnservice",
    },
    743000: {
        "branche": "Oversættelse og tolkning",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MC",
        "grp36hovedafdeling": "Reklame og øvrig erhvervsservice - Reklame og øvrig erhvervsservice   ",
        "grp127kode": 74000,
        "grp127hovedafdeling": "Anden videnservice",
    },
    749010: {
        "branche": "Landbrugskonsulenter",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MC",
        "grp36hovedafdeling": "Reklame og øvrig erhvervsservice - Reklame og øvrig erhvervsservice   ",
        "grp127kode": 74000,
        "grp127hovedafdeling": "Anden videnservice",
    },
    749090: {
        "branche": "Andre liberale, videnskabelige og tekniske tjenesteydelser i.a.n.",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MC",
        "grp36hovedafdeling": "Reklame og øvrig erhvervsservice - Reklame og øvrig erhvervsservice   ",
        "grp127kode": 74000,
        "grp127hovedafdeling": "Anden videnservice",
    },
    750000: {
        "branche": "Dyrlæger",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "M",
        "grp19hovedafdeling": "Videnservice",
        "grp36kode": "MC",
        "grp36hovedafdeling": "Reklame og øvrig erhvervsservice - Reklame og øvrig erhvervsservice   ",
        "grp127kode": 75000,
        "grp127hovedafdeling": "Dyrlæger",
    },
    771100: {
        "branche": "Udlejning og leasing af biler og lette motorkøretøjer",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 77000,
        "grp127hovedafdeling": "Udlejning og leasing af materiel",
    },
    771200: {
        "branche": "Udlejning og leasing af lastbiler",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 77000,
        "grp127hovedafdeling": "Udlejning og leasing af materiel",
    },
    772100: {
        "branche": "Udlejning og leasing af varer til fritid og sport",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 77000,
        "grp127hovedafdeling": "Udlejning og leasing af materiel",
    },
    772200: {
        "branche": "Udlejning af videobånd og videodisks",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 77000,
        "grp127hovedafdeling": "Udlejning og leasing af materiel",
    },
    772900: {
        "branche": "Udlejning og leasing af andre varer til personlig brug og husholdningsbrug i.a.n.",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 77000,
        "grp127hovedafdeling": "Udlejning og leasing af materiel",
    },
    773100: {
        "branche": "Udlejning og leasing af landbrugsmaskiner og -udstyr",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 77000,
        "grp127hovedafdeling": "Udlejning og leasing af materiel",
    },
    773200: {
        "branche": "Udlejning og leasing af entreprenørmateriel",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 77000,
        "grp127hovedafdeling": "Udlejning og leasing af materiel",
    },
    773300: {
        "branche": "Udlejning af kontormaskiner og -udstyr, computere og it-udstyr",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 77000,
        "grp127hovedafdeling": "Udlejning og leasing af materiel",
    },
    773400: {
        "branche": "Udlejning og leasing af skibe og både",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 77000,
        "grp127hovedafdeling": "Udlejning og leasing af materiel",
    },
    773500: {
        "branche": "Udlejning og leasing af luftfartøjer",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 77000,
        "grp127hovedafdeling": "Udlejning og leasing af materiel",
    },
    773900: {
        "branche": "Udlejning og leasing af andet materiel, udstyr og andre materielle aktiver i.a.n.",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 77000,
        "grp127hovedafdeling": "Udlejning og leasing af materiel",
    },
    774000: {
        "branche": "Leasing af intellektuelle ejendomsrettigheder og lignende, dog ikke ophavsretsbeskyttede værker",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 77000,
        "grp127hovedafdeling": "Udlejning og leasing af materiel",
    },
    781000: {
        "branche": "Arbejdsformidlingskontorer",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 78000,
        "grp127hovedafdeling": "Arbejdsformidling og vikarbureauer",
    },
    782000: {
        "branche": "Vikarbureauer",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 78000,
        "grp127hovedafdeling": "Arbejdsformidling og vikarbureauer",
    },
    783000: {
        "branche": "Anden personaleformidling",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 78000,
        "grp127hovedafdeling": "Arbejdsformidling og vikarbureauer",
    },
    791100: {
        "branche": "Rejsebureauer",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 79000,
        "grp127hovedafdeling": "Rejsebureauer",
    },
    791200: {
        "branche": "Rejsearrangører",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 79000,
        "grp127hovedafdeling": "Rejsebureauer",
    },
    799000: {
        "branche": "Andre reservationstjenesteydelser og tjenesteydelser i forbindelse hermed",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 79000,
        "grp127hovedafdeling": "Rejsebureauer",
    },
    801000: {
        "branche": "Private vagt- og sikkerhedstjenester",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 80000,
        "grp127hovedafdeling": "Vagt og sikkerhedstjeneste",
    },
    802000: {
        "branche": "Serviceydelser i forbindelse med sikkerhedssystemer",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 80000,
        "grp127hovedafdeling": "Vagt og sikkerhedstjeneste",
    },
    803000: {
        "branche": "Overvågning",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 80000,
        "grp127hovedafdeling": "Vagt og sikkerhedstjeneste",
    },
    811000: {
        "branche": "Kombinerede serviceydelser",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 81000,
        "grp127hovedafdeling": "Ejendomsservice, rengøring og anlægsgartnere",
    },
    812100: {
        "branche": "Almindelig rengøring i bygninger",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 81000,
        "grp127hovedafdeling": "Ejendomsservice, rengøring og anlægsgartnere",
    },
    812210: {
        "branche": "Vinduespolering",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 81000,
        "grp127hovedafdeling": "Ejendomsservice, rengøring og anlægsgartnere",
    },
    812220: {
        "branche": "Skorstensfejning",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 81000,
        "grp127hovedafdeling": "Ejendomsservice, rengøring og anlægsgartnere",
    },
    812290: {
        "branche": "Anden rengøring af bygninger og rengøring af erhvervslokaler",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 81000,
        "grp127hovedafdeling": "Ejendomsservice, rengøring og anlægsgartnere",
    },
    812900: {
        "branche": "Andre rengøringsydelser",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 81000,
        "grp127hovedafdeling": "Ejendomsservice, rengøring og anlægsgartnere",
    },
    813000: {
        "branche": "Landskabspleje",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 81000,
        "grp127hovedafdeling": "Ejendomsservice, rengøring og anlægsgartnere",
    },
    821100: {
        "branche": "Kombinerede administrationsserviceydelser",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 82000,
        "grp127hovedafdeling": "Anden operationel service",
    },
    821900: {
        "branche": "Fotokopiering, dokumentbehandling og anden specialiseret kontorservice",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 82000,
        "grp127hovedafdeling": "Anden operationel service",
    },
    822000: {
        "branche": "Call centres virksomhed",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 82000,
        "grp127hovedafdeling": "Anden operationel service",
    },
    823000: {
        "branche": "Organisering af kongresser, messer og udstillinger",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 82000,
        "grp127hovedafdeling": "Anden operationel service",
    },
    829100: {
        "branche": "Inkassovirksomhed og kreditoplysning",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 82000,
        "grp127hovedafdeling": "Anden operationel service",
    },
    829200: {
        "branche": "Pakkerier",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 82000,
        "grp127hovedafdeling": "Anden operationel service",
    },
    829900: {
        "branche": "Anden forretningsservice i.a.n.",
        "grp10kode": 8,
        "grp10hovedafdeling": "Erhvervsservice",
        "grp19kode": "N",
        "grp19hovedafdeling": "Rejsebureauer, rengøring og anden operationel service",
        "grp36kode": "N",
        "grp36hovedafdeling": "Rejsebureauer, rengøring og anden operationel service - Rejsebureauer, rengøring og anden operationel service   ",
        "grp127kode": 82000,
        "grp127hovedafdeling": "Anden operationel service",
    },
    841100: {
        "branche": "Generelle offentlige tjenester",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "O",
        "grp19hovedafdeling": "Offentlig administration, forsvar og politi",
        "grp36kode": "O",
        "grp36hovedafdeling": "Offentlig administration, forsvar og politi - Offentlig administration, forsvar og politi   ",
        "grp127kode": 84001,
        "grp127hovedafdeling": "Offentlig administration",
    },
    841200: {
        "branche": "Administration af sundhedsvæsen, undervisning, kultur og sociale forhold undtagen social sikring",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "O",
        "grp19hovedafdeling": "Offentlig administration, forsvar og politi",
        "grp36kode": "O",
        "grp36hovedafdeling": "Offentlig administration, forsvar og politi - Offentlig administration, forsvar og politi   ",
        "grp127kode": 84001,
        "grp127hovedafdeling": "Offentlig administration",
    },
    841300: {
        "branche": "Administration af og bidrag til erhvervsfremme",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "O",
        "grp19hovedafdeling": "Offentlig administration, forsvar og politi",
        "grp36kode": "O",
        "grp36hovedafdeling": "Offentlig administration, forsvar og politi - Offentlig administration, forsvar og politi   ",
        "grp127kode": 84001,
        "grp127hovedafdeling": "Offentlig administration",
    },
    842100: {
        "branche": "Udenrigsanliggender",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "O",
        "grp19hovedafdeling": "Offentlig administration, forsvar og politi",
        "grp36kode": "O",
        "grp36hovedafdeling": "Offentlig administration, forsvar og politi - Offentlig administration, forsvar og politi   ",
        "grp127kode": 84002,
        "grp127hovedafdeling": "Forsvar, politi og retsvæsen mv.",
    },
    842200: {
        "branche": "Forsvar",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "O",
        "grp19hovedafdeling": "Offentlig administration, forsvar og politi",
        "grp36kode": "O",
        "grp36hovedafdeling": "Offentlig administration, forsvar og politi - Offentlig administration, forsvar og politi   ",
        "grp127kode": 84002,
        "grp127hovedafdeling": "Forsvar, politi og retsvæsen mv.",
    },
    842300: {
        "branche": "Domstole og fængselsvæsen",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "O",
        "grp19hovedafdeling": "Offentlig administration, forsvar og politi",
        "grp36kode": "O",
        "grp36hovedafdeling": "Offentlig administration, forsvar og politi - Offentlig administration, forsvar og politi   ",
        "grp127kode": 84002,
        "grp127hovedafdeling": "Forsvar, politi og retsvæsen mv.",
    },
    842400: {
        "branche": "Politi",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "O",
        "grp19hovedafdeling": "Offentlig administration, forsvar og politi",
        "grp36kode": "O",
        "grp36hovedafdeling": "Offentlig administration, forsvar og politi - Offentlig administration, forsvar og politi   ",
        "grp127kode": 84002,
        "grp127hovedafdeling": "Forsvar, politi og retsvæsen mv.",
    },
    842500: {
        "branche": "Brandvæsen",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "O",
        "grp19hovedafdeling": "Offentlig administration, forsvar og politi",
        "grp36kode": "O",
        "grp36hovedafdeling": "Offentlig administration, forsvar og politi - Offentlig administration, forsvar og politi   ",
        "grp127kode": 84002,
        "grp127hovedafdeling": "Forsvar, politi og retsvæsen mv.",
    },
    843000: {
        "branche": "Lovpligtig socialsikring mv.",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "O",
        "grp19hovedafdeling": "Offentlig administration, forsvar og politi",
        "grp36kode": "O",
        "grp36hovedafdeling": "Offentlig administration, forsvar og politi - Offentlig administration, forsvar og politi   ",
        "grp127kode": 84001,
        "grp127hovedafdeling": "Offentlig administration",
    },
    851000: {
        "branche": "Førskoleundervisning",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "P",
        "grp19hovedafdeling": "Undervisning",
        "grp36kode": "P",
        "grp36hovedafdeling": "Undervisning - Undervisning   ",
        "grp127kode": 85001,
        "grp127hovedafdeling": "Grundskoler",
    },
    852010: {
        "branche": "Folkeskoler o.lign.",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "P",
        "grp19hovedafdeling": "Undervisning",
        "grp36kode": "P",
        "grp36hovedafdeling": "Undervisning - Undervisning   ",
        "grp127kode": 85001,
        "grp127hovedafdeling": "Grundskoler",
    },
    852020: {
        "branche": "Specialskoler for handicappede",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "P",
        "grp19hovedafdeling": "Undervisning",
        "grp36kode": "P",
        "grp36hovedafdeling": "Undervisning - Undervisning   ",
        "grp127kode": 85001,
        "grp127hovedafdeling": "Grundskoler",
    },
    853110: {
        "branche": "Ungdoms- og efterskoler",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "P",
        "grp19hovedafdeling": "Undervisning",
        "grp36kode": "P",
        "grp36hovedafdeling": "Undervisning - Undervisning   ",
        "grp127kode": 85002,
        "grp127hovedafdeling": "Gymnasier og erhvervsfaglige skoler",
    },
    853120: {
        "branche": "Gymnasier, studenter- og HF-kurser",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "P",
        "grp19hovedafdeling": "Undervisning",
        "grp36kode": "P",
        "grp36hovedafdeling": "Undervisning - Undervisning   ",
        "grp127kode": 85002,
        "grp127hovedafdeling": "Gymnasier og erhvervsfaglige skoler",
    },
    853200: {
        "branche": "Tekniske skoler og fagskoler",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "P",
        "grp19hovedafdeling": "Undervisning",
        "grp36kode": "P",
        "grp36hovedafdeling": "Undervisning - Undervisning   ",
        "grp127kode": 85002,
        "grp127hovedafdeling": "Gymnasier og erhvervsfaglige skoler",
    },
    854100: {
        "branche": "Videregående uddannelser ikke på universitetsniveau",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "P",
        "grp19hovedafdeling": "Undervisning",
        "grp36kode": "P",
        "grp36hovedafdeling": "Undervisning - Undervisning   ",
        "grp127kode": 85003,
        "grp127hovedafdeling": "Videregående uddannelsesinstitutioner",
    },
    854200: {
        "branche": "Videregående uddannelser på universitetsniveau",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "P",
        "grp19hovedafdeling": "Undervisning",
        "grp36kode": "P",
        "grp36hovedafdeling": "Undervisning - Undervisning   ",
        "grp127kode": 85003,
        "grp127hovedafdeling": "Videregående uddannelsesinstitutioner",
    },
    855100: {
        "branche": "Undervisning inden for sport og fritid",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "P",
        "grp19hovedafdeling": "Undervisning",
        "grp36kode": "P",
        "grp36hovedafdeling": "Undervisning - Undervisning   ",
        "grp127kode": 85004,
        "grp127hovedafdeling": "Voksenundervisning mv.",
    },
    855200: {
        "branche": "Undervisning i kulturelle discipliner",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "P",
        "grp19hovedafdeling": "Undervisning",
        "grp36kode": "P",
        "grp36hovedafdeling": "Undervisning - Undervisning   ",
        "grp127kode": 85004,
        "grp127hovedafdeling": "Voksenundervisning mv.",
    },
    855300: {
        "branche": "Køreskoler",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "P",
        "grp19hovedafdeling": "Undervisning",
        "grp36kode": "P",
        "grp36hovedafdeling": "Undervisning - Undervisning   ",
        "grp127kode": 85004,
        "grp127hovedafdeling": "Voksenundervisning mv.",
    },
    855900: {
        "branche": "Anden undervisning i.a.n.",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "P",
        "grp19hovedafdeling": "Undervisning",
        "grp36kode": "P",
        "grp36hovedafdeling": "Undervisning - Undervisning   ",
        "grp127kode": 85004,
        "grp127hovedafdeling": "Voksenundervisning mv.",
    },
    856000: {
        "branche": "Hjælpeydelser i forbindelse med undervisning",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "P",
        "grp19hovedafdeling": "Undervisning",
        "grp36kode": "P",
        "grp36hovedafdeling": "Undervisning - Undervisning   ",
        "grp127kode": 85004,
        "grp127hovedafdeling": "Voksenundervisning mv.",
    },
    861000: {
        "branche": "Hospitaler",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QA",
        "grp36hovedafdeling": "Sundhedsvæsen - Sundhedsvæsen   ",
        "grp127kode": 86001,
        "grp127hovedafdeling": "Hospitaler",
    },
    862100: {
        "branche": "Alment praktiserende læger",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QA",
        "grp36hovedafdeling": "Sundhedsvæsen - Sundhedsvæsen   ",
        "grp127kode": 86002,
        "grp127hovedafdeling": "Læger, tandlæger mv.",
    },
    862200: {
        "branche": "Praktiserende speciallæger",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QA",
        "grp36hovedafdeling": "Sundhedsvæsen - Sundhedsvæsen   ",
        "grp127kode": 86002,
        "grp127hovedafdeling": "Læger, tandlæger mv.",
    },
    862300: {
        "branche": "Praktiserende tandlæger",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QA",
        "grp36hovedafdeling": "Sundhedsvæsen - Sundhedsvæsen   ",
        "grp127kode": 86002,
        "grp127hovedafdeling": "Læger, tandlæger mv.",
    },
    869010: {
        "branche": "Sundhedspleje, hjemmesygepleje og jordemødre mv.",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QA",
        "grp36hovedafdeling": "Sundhedsvæsen - Sundhedsvæsen   ",
        "grp127kode": 86002,
        "grp127hovedafdeling": "Læger, tandlæger mv.",
    },
    869020: {
        "branche": "Fysio- og ergoterapeuter",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QA",
        "grp36hovedafdeling": "Sundhedsvæsen - Sundhedsvæsen   ",
        "grp127kode": 86002,
        "grp127hovedafdeling": "Læger, tandlæger mv.",
    },
    869030: {
        "branche": "Psykologisk rådgivning",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QA",
        "grp36hovedafdeling": "Sundhedsvæsen - Sundhedsvæsen   ",
        "grp127kode": 86002,
        "grp127hovedafdeling": "Læger, tandlæger mv.",
    },
    869040: {
        "branche": "Kiropraktorer",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QA",
        "grp36hovedafdeling": "Sundhedsvæsen - Sundhedsvæsen   ",
        "grp127kode": 86002,
        "grp127hovedafdeling": "Læger, tandlæger mv.",
    },
    869090: {
        "branche": "Sundhedsvæsen i øvrigt i.a.n.",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QA",
        "grp36hovedafdeling": "Sundhedsvæsen - Sundhedsvæsen   ",
        "grp127kode": 86002,
        "grp127hovedafdeling": "Læger, tandlæger mv.",
    },
    871010: {
        "branche": "Plejehjem",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 87000,
        "grp127hovedafdeling": "Plejehjem mv.",
    },
    871020: {
        "branche": "Institutionsophold med sygepleje i.a.n.",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 87000,
        "grp127hovedafdeling": "Plejehjem mv.",
    },
    872010: {
        "branche": "Døgninstitutioner for personer med psykiske handicap",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 87000,
        "grp127hovedafdeling": "Plejehjem mv.",
    },
    872020: {
        "branche": "Behandlingshjem for stofmisbrugere og alkoholskadede",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 87000,
        "grp127hovedafdeling": "Plejehjem mv.",
    },
    873010: {
        "branche": "Døgninstitutioner for personer med fysisk handicap",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 87000,
        "grp127hovedafdeling": "Plejehjem mv.",
    },
    873020: {
        "branche": "Beskyttede boliger o.l.",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 87000,
        "grp127hovedafdeling": "Plejehjem mv.",
    },
    879010: {
        "branche": "Døgninstitutioner for børn og unge",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 87000,
        "grp127hovedafdeling": "Plejehjem mv.",
    },
    879020: {
        "branche": "Familiepleje",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 87000,
        "grp127hovedafdeling": "Plejehjem mv.",
    },
    879090: {
        "branche": "Andre former for institutionsophold",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 87000,
        "grp127hovedafdeling": "Plejehjem mv.",
    },
    881010: {
        "branche": "Hjemmehjælp",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 88000,
        "grp127hovedafdeling": "Daginstitutioner og dagcentre mv. - Daginstitutioner og dagcentre mv. ",
    },
    881020: {
        "branche": "Dagcentre mv.",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 88000,
        "grp127hovedafdeling": "Daginstitutioner og dagcentre mv. - Daginstitutioner og dagcentre mv. ",
    },
    881030: {
        "branche": "Revalideringsinstitutioner",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 88000,
        "grp127hovedafdeling": "Daginstitutioner og dagcentre mv. - Daginstitutioner og dagcentre mv. ",
    },
    889110: {
        "branche": "Dagplejemødre",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 88000,
        "grp127hovedafdeling": "Daginstitutioner og dagcentre mv. - Daginstitutioner og dagcentre mv. ",
    },
    889120: {
        "branche": "Vuggestuer",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 88000,
        "grp127hovedafdeling": "Daginstitutioner og dagcentre mv. - Daginstitutioner og dagcentre mv. ",
    },
    889130: {
        "branche": "Børnehaver",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 88000,
        "grp127hovedafdeling": "Daginstitutioner og dagcentre mv. - Daginstitutioner og dagcentre mv. ",
    },
    889140: {
        "branche": "Skolefritidsordninger og fritidshjem",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 88000,
        "grp127hovedafdeling": "Daginstitutioner og dagcentre mv. - Daginstitutioner og dagcentre mv. ",
    },
    889150: {
        "branche": "Aldersintegrerede institutioner",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 88000,
        "grp127hovedafdeling": "Daginstitutioner og dagcentre mv. - Daginstitutioner og dagcentre mv. ",
    },
    889160: {
        "branche": "Fritids- og ungdomsklubber",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 88000,
        "grp127hovedafdeling": "Daginstitutioner og dagcentre mv. - Daginstitutioner og dagcentre mv. ",
    },
    889910: {
        "branche": "Foreninger, legater og fonde med sygdomsbekæmpende, sociale og velgørende formål",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 88000,
        "grp127hovedafdeling": "Daginstitutioner og dagcentre mv. - Daginstitutioner og dagcentre mv. ",
    },
    889920: {
        "branche": "Flygtninge- og asylcentre",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 88000,
        "grp127hovedafdeling": "Daginstitutioner og dagcentre mv. - Daginstitutioner og dagcentre mv. ",
    },
    889990: {
        "branche": "Andre sociale foranstaltninger uden institutionsophold i.a.n.",
        "grp10kode": 9,
        "grp10hovedafdeling": "Offentlig administration, undervisning og sundhed",
        "grp19kode": "Q",
        "grp19hovedafdeling": "Sundhed og socialvæsen",
        "grp36kode": "QB",
        "grp36hovedafdeling": "Sociale institutioner - Sociale institutioner   ",
        "grp127kode": 88000,
        "grp127hovedafdeling": "Daginstitutioner og dagcentre mv. - Daginstitutioner og dagcentre mv. ",
    },
    900110: {
        "branche": "Teater- og koncertvirksomhed",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 90000,
        "grp127hovedafdeling": "Teater, musik og kunst",
    },
    900120: {
        "branche": "Selvstændigt udøvende scenekunstnere",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 90000,
        "grp127hovedafdeling": "Teater, musik og kunst",
    },
    900200: {
        "branche": "Hjælpeaktiviteter i forbindelse med scenekunst",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 90000,
        "grp127hovedafdeling": "Teater, musik og kunst",
    },
    900300: {
        "branche": "Kunstnerisk skaben",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 90000,
        "grp127hovedafdeling": "Teater, musik og kunst",
    },
    900400: {
        "branche": "Drift af teater- og koncertsale, kulturhuse mv.",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 90000,
        "grp127hovedafdeling": "Teater, musik og kunst",
    },
    910110: {
        "branche": "Biblioteker",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 91000,
        "grp127hovedafdeling": "Biblioteker, museer mv.",
    },
    910120: {
        "branche": "Arkiver",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 91000,
        "grp127hovedafdeling": "Biblioteker, museer mv.",
    },
    910200: {
        "branche": "Museer",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 91000,
        "grp127hovedafdeling": "Biblioteker, museer mv.",
    },
    910300: {
        "branche": "Historiske monumenter og bygninger og lignende attraktioner",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 91000,
        "grp127hovedafdeling": "Biblioteker, museer mv.",
    },
    910400: {
        "branche": "Botaniske og zoologiske haver samt naturreservater",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 91000,
        "grp127hovedafdeling": "Biblioteker, museer mv.",
    },
    920000: {
        "branche": "Lotteri- og anden spillevirksomhed",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 92000,
        "grp127hovedafdeling": "Lotteri og andet spil",
    },
    931100: {
        "branche": "Drift af sportsanlæg",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 93001,
        "grp127hovedafdeling": "Sportsklubber",
    },
    931200: {
        "branche": "Sportsklubber",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 93001,
        "grp127hovedafdeling": "Sportsklubber",
    },
    931300: {
        "branche": "Fitnesscentre",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 93001,
        "grp127hovedafdeling": "Sportsklubber",
    },
    931900: {
        "branche": "Andre sportsaktiviteter",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 93001,
        "grp127hovedafdeling": "Sportsklubber",
    },
    932100: {
        "branche": "Forlystelsesparker o.l.",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 93002,
        "grp127hovedafdeling": "Forlystelsesparker og andre fritidsaktiviteter",
    },
    932910: {
        "branche": "Lystbådehavne",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 93002,
        "grp127hovedafdeling": "Forlystelsesparker og andre fritidsaktiviteter",
    },
    932990: {
        "branche": "Andre forlystelser og fritidsaktiviteter",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "R",
        "grp19hovedafdeling": "Kultur og fritid",
        "grp36kode": "R",
        "grp36hovedafdeling": "Kultur og fritid - Kultur og fritid   ",
        "grp127kode": 93002,
        "grp127hovedafdeling": "Forlystelsesparker og andre fritidsaktiviteter",
    },
    941100: {
        "branche": "Erhvervs- og arbejdsgiverorganisationer",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 94000,
        "grp127hovedafdeling": "Organisationer og foreninger",
    },
    941200: {
        "branche": "Faglige sammenslutninger",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 94000,
        "grp127hovedafdeling": "Organisationer og foreninger",
    },
    942000: {
        "branche": "Fagforeninger",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 94000,
        "grp127hovedafdeling": "Organisationer og foreninger",
    },
    949100: {
        "branche": "Religiøse institutioner og foreninger",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 94000,
        "grp127hovedafdeling": "Organisationer og foreninger",
    },
    949200: {
        "branche": "Politiske partier",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 94000,
        "grp127hovedafdeling": "Organisationer og foreninger",
    },
    949900: {
        "branche": "Andre organisationer og foreninger i.a.n.",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 94000,
        "grp127hovedafdeling": "Organisationer og foreninger",
    },
    951100: {
        "branche": "Reparation af computere og ydre enheder",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 95000,
        "grp127hovedafdeling": "Reparation af husholdningsudstyr",
    },
    951200: {
        "branche": "Reparation af kommunikationsudstyr",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 95000,
        "grp127hovedafdeling": "Reparation af husholdningsudstyr",
    },
    952100: {
        "branche": "Reparation af forbrugerelektronik",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 95000,
        "grp127hovedafdeling": "Reparation af husholdningsudstyr",
    },
    952200: {
        "branche": "Reparation af husholdningsapparater og redskaber til hus og have",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 95000,
        "grp127hovedafdeling": "Reparation af husholdningsudstyr",
    },
    952300: {
        "branche": "Reparation af skotøj og lædervarer",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 95000,
        "grp127hovedafdeling": "Reparation af husholdningsudstyr",
    },
    952400: {
        "branche": "Reparation af møbler og boligudstyr",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 95000,
        "grp127hovedafdeling": "Reparation af husholdningsudstyr",
    },
    952500: {
        "branche": "Reparation af ure, smykker og guld- og sølvvarer",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 95000,
        "grp127hovedafdeling": "Reparation af husholdningsudstyr",
    },
    952900: {
        "branche": "Reparation af andre varer til personligt brug og husholdningsbrug",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 95000,
        "grp127hovedafdeling": "Reparation af husholdningsudstyr",
    },
    960110: {
        "branche": "Erhvervs- og institutionsvaskerier",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 96000,
        "grp127hovedafdeling": "Frisører, vaskerier og andre serviceydelser",
    },
    960120: {
        "branche": "Renserier, selvbetjeningsvaskerier mv.",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 96000,
        "grp127hovedafdeling": "Frisører, vaskerier og andre serviceydelser",
    },
    960210: {
        "branche": "Frisørsaloner",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 96000,
        "grp127hovedafdeling": "Frisører, vaskerier og andre serviceydelser",
    },
    960220: {
        "branche": "Skønheds- og hudpleje",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 96000,
        "grp127hovedafdeling": "Frisører, vaskerier og andre serviceydelser",
    },
    960300: {
        "branche": "Bedemænd og begravelsesvæsen",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 96000,
        "grp127hovedafdeling": "Frisører, vaskerier og andre serviceydelser",
    },
    960400: {
        "branche": "Aktiviteter vedrørende fysisk velvære",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 96000,
        "grp127hovedafdeling": "Frisører, vaskerier og andre serviceydelser",
    },
    960900: {
        "branche": "Andre personlige serviceydelser i.a.n.",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 96000,
        "grp127hovedafdeling": "Frisører, vaskerier og andre serviceydelser",
    },
    970000: {
        "branche": "Husholdninger med ansat medhjælp",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 97000,
        "grp127hovedafdeling": "Private husholdninger med ansat medhjælp - Private husholdninger med ansat medhjælp ",
    },
    981000: {
        "branche": "Private husholdningers produktion af varer til eget brug, i.a.n.",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 97000,
        "grp127hovedafdeling": "Private husholdninger med ansat medhjælp - Private husholdninger med ansat medhjælp ",
    },
    982000: {
        "branche": "Private husholdningers produktion af tjenesteydelser til eget brug, i.a.n.",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 97000,
        "grp127hovedafdeling": "Private husholdninger med ansat medhjælp - Private husholdninger med ansat medhjælp ",
    },
    990000: {
        "branche": "Eksterritoriale organisationer og organer",
        "grp10kode": 10,
        "grp10hovedafdeling": "Kultur, fritid og anden service",
        "grp19kode": "S",
        "grp19hovedafdeling": "Andre serviceydelser mv",
        "grp36kode": "S",
        "grp36hovedafdeling": "Andre serviceydelser  mv.",
        "grp127kode": 99000,
        "grp127hovedafdeling": "Eksterritoriale organisationer og organer",
    },
}